from __future__ import annotations

import asyncio
import io
import os
import pkgutil
import zipfile
from typing import Any, Tuple, Callable, TypeVar, Dict

import boto3

import meadowrun.aws_integration.management_lambdas.adjust_ec2_instances
import meadowrun.aws_integration.management_lambdas.clean_up
from meadowrun.aws_integration.aws_core import (
    _get_account_number,
    _get_default_region_name,
)
from meadowrun.aws_integration.aws_permissions_install import _MANAGEMENT_LAMBDA_ROLE
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    ignore_boto3_error_code,
)
from meadowrun.shared import create_zipfile

_T = TypeVar("_T")

# the name of the lambda that runs adjust_ec2_instances.py
_EC2_ALLOC_LAMBDA_NAME = "meadowrun_ec2_alloc_lambda"
# the EventBridge rule that triggers the lambda
_EC2_ALLOC_LAMBDA_SCHEDULE_RULE = "meadowrun_ec2_alloc_lambda_schedule_rule"
# the name of the lambda that runs clean_up.py
_CLEAN_UP_LAMBDA_NAME = "meadowrun_clean_up"
_CLEAN_UP_LAMBDA_SCHEDULE_RULE = "meadowrun_clean_up_lambda_schedule_rule"


def _get_zipped_lambda_code(overrides: Dict[str, str]) -> bytes:
    """
    Gets the contents of the ec2_alloc_lambda folder as a zip file. This is the code we
    want to run as a lambda.

    Warning, this doesn't recurse into any subdirectories (because it is not currently
    needed)
    """
    lambda_root_path = meadowrun.aws_integration.management_lambdas.__path__[0]
    module_names = [name for _, name, _ in pkgutil.iter_modules([lambda_root_path])]
    path_prefix = meadowrun.aws_integration.management_lambdas.__name__.replace(
        ".", os.path.sep
    )

    with io.BytesIO() as buffer:
        with create_zipfile(buffer, "w", zipfile.ZIP_DEFLATED) as zf:
            for module_name in module_names:
                if module_name == "config" and overrides:
                    with open(
                        os.path.join(lambda_root_path, f"{module_name}.py"),
                        "r",
                        encoding="utf-8",
                    ) as orig_config:
                        result = []
                        for line in orig_config:
                            for key, value in overrides.items():
                                if line.startswith(f"{key} = "):
                                    result.append(f"{key} = {value}\n")
                                else:
                                    result.append(line)
                        zf.writestr(
                            os.path.join(path_prefix, module_name + ".py"),
                            "".join(result),
                        )
                else:
                    zf.write(
                        os.path.join(lambda_root_path, module_name + ".py"),
                        os.path.join(path_prefix, module_name + ".py"),
                    )

        buffer.seek(0)

        return buffer.read()


async def _create_management_lambda(
    lambda_client: Any,
    lambda_handler: Any,
    lambda_name: str,
    schedule_rule_name: str,
    schedule_expression: str,
    region_name: str,
    overrides: Dict[str, str],
) -> None:
    """Creates the ec2 alloc lambda assuming it does not already exist"""
    account_number = _get_account_number()

    # create the lambda
    def create_function_if_not_exists() -> Tuple[bool, None]:
        ignore_boto3_error_code(
            lambda: lambda_client.create_function(
                FunctionName=lambda_name,
                Runtime="python3.9",
                Role=f"arn:aws:iam::{account_number}:role/{_MANAGEMENT_LAMBDA_ROLE}",
                Handler=f"{lambda_handler.__module__}.{lambda_handler.__name__}",
                Code={"ZipFile": _get_zipped_lambda_code(overrides)},
                Timeout=120,
                MemorySize=128,  # memory available in MB
            ),
            "ResourceConflictException",
        )
        return True, None

    # totally crazy, but sometimes you just have to wait 5-10 seconds after
    # creating the role to be able to create a lambda with that role:
    # https://stackoverflow.com/a/37438525
    await _retry(
        lambda: ignore_boto3_error_code(
            create_function_if_not_exists, "InvalidParameterValueException"
        ),
        10,
        2,
        "Waiting for newly created AWS IAM role to become available...",
    )

    # now create an EventBridge rule that triggers every 1 minute
    events_client = boto3.client("events")
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/events.html#EventBridge.Client.put_rule
    events_client.put_rule(
        Name=schedule_rule_name, ScheduleExpression=schedule_expression
    )

    # add the lambda as a target for that rule
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/events.html#EventBridge.Client.put_targets
    events_client.put_targets(
        Rule=schedule_rule_name,
        Targets=[
            {
                "Id": lambda_name,
                "Arn": f"arn:aws:lambda:{region_name}:{account_number}:function:"
                f"{lambda_name}",
            }
        ],
    )

    # add permissions for that rule to invoke this lambda
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/lambda.html#Lambda.Client.add_permission
    ignore_boto3_error_code(
        lambda: lambda_client.add_permission(
            FunctionName=lambda_name,
            StatementId=f"{schedule_rule_name}_invokes_{lambda_name}",
            Action="lambda:InvokeFunction",
            Principal="events.amazonaws.com",
            SourceArn=(
                f"arn:aws:events:{region_name}:{account_number}:rule/"
                f"{schedule_rule_name}"
            ),
        ),
        "ResourceConflictException",
    )


def _set_retention_policy_lambda_logs(lambda_name: str, region_name: str) -> None:
    """
    Cloudwatch log groups are automatically created for lambdas when they are first run
    with an infinite retention period. This function preemptively creates the log group
    so that we can set the retention period.
    """
    logs_client = boto3.client("logs", region_name=region_name)
    ignore_boto3_error_code(
        lambda: logs_client.create_log_group(logGroupName=f"/aws/lambda/{lambda_name}"),
        "ResourceAlreadyExistsException",
    )
    logs_client.put_retention_policy(
        logGroupName=f"/aws/lambda/{lambda_name}", retentionInDays=3
    )


async def _ensure_management_lambda(
    lambda_handler: Any,
    lambda_name: str,
    schedule_rule_name: str,
    schedule_expression: str,
    update_if_exists: bool,
    overrides: Dict[str, str],
) -> None:
    """
    Create the specified management lambda if it doesn't exist. If update_if_exists is
    true, updates the code if the lambda already exists.

    Even if this is called with update_if_exists, it is not guaranteed to update the
    code if another process creates the lambda after this function starts executing.

    The meadowrun management lambda role must exist already
    """

    region_name = await _get_default_region_name()
    lambda_client = boto3.client("lambda", region_name=region_name)

    exists, _ = ignore_boto3_error_code(
        lambda: lambda_client.get_function(FunctionName=lambda_name),
        "ResourceNotFoundException",
    )

    if not exists:
        await _create_management_lambda(
            lambda_client,
            lambda_handler,
            lambda_name,
            schedule_rule_name,
            schedule_expression,
            region_name,
            overrides,
        )
        _set_retention_policy_lambda_logs(lambda_name, region_name)
    elif update_if_exists:
        lambda_client.update_function_code(
            FunctionName=lambda_name, ZipFile=_get_zipped_lambda_code(overrides)
        )
        _set_retention_policy_lambda_logs(lambda_name, region_name)


async def ensure_ec2_alloc_lambda(
    update_if_exists: bool, overrides: Dict[str, str]
) -> None:
    await _ensure_management_lambda(
        meadowrun.aws_integration.management_lambdas.adjust_ec2_instances.lambda_handler,  # noqa: E501
        _EC2_ALLOC_LAMBDA_NAME,
        _EC2_ALLOC_LAMBDA_SCHEDULE_RULE,
        "rate(1 minute)",
        update_if_exists,
        overrides,
    )


async def ensure_clean_up_lambda(
    update_if_exists: bool, overrides: Dict[str, str]
) -> None:
    await _ensure_management_lambda(
        meadowrun.aws_integration.management_lambdas.clean_up.lambda_handler,
        _CLEAN_UP_LAMBDA_NAME,
        _CLEAN_UP_LAMBDA_SCHEDULE_RULE,
        "rate(3 hours)",
        update_if_exists,
        overrides,
    )


async def _retry(
    function: Callable[[], Tuple[bool, _T]],
    max_num_attempts: int = 3,
    delay_seconds: float = 1,
    retry_message: str = "Retrying on error",
) -> _T:
    i = 0
    while True:
        success, result = function()
        if success:
            return result
        else:
            i += 1
            if i >= max_num_attempts:
                raise ValueError(f"Failed after {i} attempts")
            else:
                print(retry_message)
                await asyncio.sleep(delay_seconds)



"""
This contains "core" functionality used by more than one aspect of Meadowrun when
interacting with Azure
"""

from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING, Optional, Tuple, Set, Any

if TYPE_CHECKING:
    from typing_extensions import Literal

import aiohttp

from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
    azure_rest_api_paged,
    azure_rest_api_poll,
    get_subscription_id,
    wait_for_poll,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_identity import (
    IMDS_AUTHORITY,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
    ResourceExistsError,
    ResourceNotFoundError,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api import (
    StorageAccount,
    azure_table_api,
    table_key_url,
)
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    LAST_USED_TABLE_NAME,
    MEADOWRUN_RESOURCE_GROUP_NAME,
    RESOURCE_TYPES_TYPE,
)


def get_default_location() -> str:
    # TODO try `az config get defaults.location`. Then our own custom config, maybe?
    return "eastus"


_MEADOWRUN_RESOURCE_GROUP_ENSURED = False


async def ensure_meadowrun_resource_group(location: str) -> str:
    """
    Creates the meadowrun resource group if it doesn't already exist. This resource
    group will contain all meadowrun-generated resources. Returns the path to the
    resource group.
    """
    global _MEADOWRUN_RESOURCE_GROUP_ENSURED

    subscription_id = await get_subscription_id()
    resource_group_path = (
        f"subscriptions/{subscription_id}/resourcegroups/"
        f"{MEADOWRUN_RESOURCE_GROUP_NAME}"
    )

    if not _MEADOWRUN_RESOURCE_GROUP_ENSURED:
        try:
            await azure_rest_api("GET", resource_group_path, "2021-04-01")
        except ResourceNotFoundError:
            print(
                f"The meadowrun resource group ({MEADOWRUN_RESOURCE_GROUP_NAME}) "
                f"doesn't exist, creating it now in {location}"
            )
            await azure_rest_api(
                "PUT",
                resource_group_path,
                "2021-04-01",
                json_content={"location": location},
            )

        _MEADOWRUN_RESOURCE_GROUP_ENSURED = True

    return resource_group_path


async def get_current_vm_name() -> Optional[str]:
    async with aiohttp.request(
        "GET",
        f"{IMDS_AUTHORITY}/metadata/instance/compute/name"
        "?api-version=2021-02-01&format=text",
        headers={"Metadata": "true"},
    ) as response:
        if not response.ok:
            return None

        return await response.text()


async def get_current_ip_address_on_vm() -> Optional[str]:
    """
    Assuming we're running on an Azure VM, get our current public ip address. If we're
    not running on an Azure VM, or we're running on an Azure VM without a public IP
    address, or if we're running on an Azure VM with a Standard SKU public IP address
    (see comments in _provision_nic_with_public_ip), we will return None.
    """
    async with aiohttp.request(
        "GET",
        f"{IMDS_AUTHORITY}/metadata/instance/network/interface/0/ipv4/ipAddress"
        "/0/publicIpAddress?api-version=2021-02-01&format=text",
        headers={"Metadata": "true"},
    ) as response:
        if not response.ok:
            return None

        return await response.text()


async def get_scheduled_events_on_vm() -> Optional[Any]:
    # See https://docs.microsoft.com/en-us/azure/virtual-machines/linux/scheduled-events
    async with aiohttp.request(
        "GET",
        f"{IMDS_AUTHORITY}/metadata/scheduledevents?api-version=2020-07-01",
        headers={"Metadata": "true"},
    ) as response:
        if not response.ok:
            return None

        return await response.json()


_MEADOWRUN_MANAGED_IDENTITY = "meadowrun-managed-identity"


async def _ensure_managed_identity(location: str) -> Tuple[str, str]:
    """Returns identity id, client id"""
    path = (
        f"{await ensure_meadowrun_resource_group(location)}/providers/"
        "Microsoft.ManagedIdentity/userAssignedIdentities/"
        f"{_MEADOWRUN_MANAGED_IDENTITY}"
    )
    try:
        identity = await azure_rest_api("GET", path, "2018-11-30")
    except ResourceNotFoundError:
        print(
            f"Azure managed identity {_MEADOWRUN_MANAGED_IDENTITY} does not exist, "
            f"creating it now"
        )

        identity = await azure_rest_api(
            "PUT", path, "2018-11-30", json_content={"location": location}
        )
        principal_id = identity["properties"]["principalId"]

        await assign_role_to_principal(
            "Contributor", principal_id, location, "ServicePrincipal"
        )
        await assign_role_to_principal(
            "Key Vault Secrets User",
            principal_id,
            location,
            "ServicePrincipal",
        )

    return identity["id"], identity["properties"]["clientId"]


async def assign_role_to_principal(
    role_name: str,
    principal_id: str,
    location: str,
    principal_type: Literal[None, "ServicePrincipal"] = None,
) -> None:
    """
    Assigns the specified role to the specified principal (e.g. user or identity) in the
    scope of the meadowrun resource group.

    principal_type should be set to ServicePrincipal as per the recommendation in
    https://docs.microsoft.com/en-us/azure/role-based-access-control/role-assignments-template#new-service-principal
    if the service principal is brand new.
    """
    subscription_id = await get_subscription_id()

    # Bizarrely, setting principal_type to ServicePrincipal for newly created service
    # identities only seems to have an effect in 2018-09-01-preview (or later according
    # to the docs), but the role_definitions property is missing on 2018-09-01-preview
    # (and later) API versions. So it seems like we need to use two different API
    # versions in this function
    roles = [
        role
        async for page in azure_rest_api_paged(
            "GET",
            f"subscriptions/{subscription_id}/providers/Microsoft.Authorization/"
            "roleDefinitions",
            "2018-01-01-preview",
            query_parameters={"$filter": f"roleName eq '{role_name}'"},
        )
        for role in page["value"]
    ]

    if len(roles) == 0:
        raise ValueError(f"Role {role_name} was not found")
    elif len(roles) > 1:
        raise ValueError(f"More than one role {role_name} was found")

    parameters = {
        "properties": {
            "roleDefinitionId": roles[0]["id"],
            "principalId": principal_id,
        }
    }
    if principal_type:
        parameters["properties"]["principalType"] = principal_type
    try:
        await azure_rest_api(
            "PUT",
            f"{await ensure_meadowrun_resource_group(location)}/providers/"
            f"Microsoft.Authorization/roleAssignments/{uuid.uuid4()}",
            "2018-09-01-preview",
            json_content=parameters,
        )
    except ResourceExistsError:
        # this means the role assignment already exists
        pass


async def ensure_meadowrun_storage_account(
    location: str, on_missing: Literal["raise", "create"]
) -> StorageAccount:
    """Returns (storage account name, key)"""
    subscription_id = await get_subscription_id()
    # the storage account name must be globally unique in Azure (across all users),
    # alphanumeric, and 24 characters or less. Our strategy is to use "mr" (for
    # meadowrun) plus the last 22 letters/numbers of the subscription id and hope for
    # the best.
    # TODO we need a way to manually set the storage account name in case this ends
    # up colliding
    storage_account_name = "mr" + subscription_id.replace("-", "")[-22:]

    storage_account_path = (
        f"{await ensure_meadowrun_resource_group(location)}/providers/"
        f"Microsoft.Storage/storageAccounts/{storage_account_name}"
    )
    # get the key to the storage account. If the storage account doesn't exist, create
    # it and then get the key
    try:
        # https://docs.microsoft.com/en-us/rest/api/storagerp/storage-accounts/list-keys
        keys = await azure_rest_api(
            "POST", f"{storage_account_path}/listKeys", "2021-09-01"
        )
    except ResourceNotFoundError:
        if on_missing == "raise":
            raise ValueError(f"Storage account {storage_account_name} does not exist")
        elif on_missing == "create":
            print(
                f"Storage account {storage_account_name} does not exist, creating it "
                f"now"
            )
            await wait_for_poll(
                await azure_rest_api_poll(
                    "PUT",
                    f"{storage_account_path}",
                    "2021-09-01",
                    "LocationStatusCode",
                    json_content={
                        "sku": {
                            # Standard (as opposed to Premium latency). Locally
                            # redundant storage (i.e. not very redundant, as opposed to
                            # zone-, geo-, or geo-and-zone- redundant storage)
                            "name": "Standard_LRS"
                        },
                        "kind": "StorageV2",
                        "location": location,
                    },
                )
            )
            keys = await azure_rest_api(
                "POST", f"{storage_account_path}/listKeys", "2021-09-01"
            )
        else:
            raise ValueError(f"Unexpected value for on_missing {on_missing}")

    return StorageAccount(
        storage_account_name, keys["keys"][0]["value"], storage_account_path
    )


_STORAGE_ACCOUNT: Optional[StorageAccount] = None
_EXISTING_TABLES: Set[str] = set()


async def ensure_table(
    table_name: str, location: str, on_missing: Literal["raise", "create"]
) -> StorageAccount:
    """
    Gets the TableClient for the specified table, creates it if it doesn't exist
    (depending on the value of the on_missing parameter). Multiple calls for the same
    table in the same process should be fast.
    """
    global _STORAGE_ACCOUNT
    if _STORAGE_ACCOUNT is None:
        # first, get the key to the storage account. If the storage account doesn't
        # exist, create it and then get the key
        _STORAGE_ACCOUNT = await ensure_meadowrun_storage_account(location, "create")

    if table_name not in _EXISTING_TABLES:
        # check if the table exists and create it if it doesn't
        table_path = (
            f"{_STORAGE_ACCOUNT.get_path()}/tableServices/default/tables/{table_name}"
        )
        try:
            await azure_rest_api("GET", table_path, "2021-09-01")
        except ResourceNotFoundError:
            if on_missing == "raise":
                raise ValueError(
                    f"Table {_STORAGE_ACCOUNT.name}/{table_name} does not exist"
                )
            elif on_missing == "create":
                print(
                    f"Table {_STORAGE_ACCOUNT.name}/{table_name} does not exist, "
                    "creating it now"
                )
                # https://docs.microsoft.com/en-us/rest/api/storagerp/table/create
                await azure_rest_api("PUT", table_path, "2021-09-01")
            else:
                raise ValueError(f"Unexpected value for on_missing {on_missing}")

        _EXISTING_TABLES.add(table_name)

    return _STORAGE_ACCOUNT


async def record_last_used(
    resource_type: RESOURCE_TYPES_TYPE, resource_name: str, location: str
) -> None:
    try:
        storage_account = await ensure_table(LAST_USED_TABLE_NAME, location, "create")
        # no need to insert timestamp, that's always included in the Azure table
        # metadata
        await azure_table_api(
            "PUT",
            storage_account,
            table_key_url(LAST_USED_TABLE_NAME, resource_type, resource_name),
            json_content={},
        )
    except Exception as e:
        logging.error(
            f"Warning, failed when trying to record usage of {resource_name}",
            exc_info=e,
        )



from __future__ import annotations

import asyncio
import dataclasses
import datetime
import functools
import json
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Callable,
    Coroutine,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

import meadowrun.azure_integration.azure_vms
from meadowrun.azure_integration.azure_meadowrun_core import (
    ensure_table,
    get_default_location,
)
from meadowrun.azure_integration.azure_ssh_keys import ensure_meadowrun_key_pair
from meadowrun.azure_integration.blob_storage import get_azure_blob_container
from meadowrun.azure_integration.grid_tasks_queue import (
    Queue,
    add_tasks,
    add_worker_shutdown_message,
    create_queues_for_job,
    get_results_unordered,
    retry_task,
    worker_function,
)
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    ALLOCATED_TIME,
    LAST_UPDATE_TIME,
    NON_CONSUMABLE_RESOURCES,
    PREVENT_FURTHER_ALLOCATION,
    PUBLIC_ADDRESS,
    RESOURCES_ALLOCATED,
    RESOURCES_AVAILABLE,
    RUNNING_JOBS,
    SINGLE_PARTITION_KEY,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
    ResourceExistsError,
    ResourceModifiedError,
    ResourceNotFoundError,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api import (
    StorageAccount,
    azure_table_api,
    azure_table_api_paged,
    table_key_url,
)
from meadowrun.azure_integration.mgmt_functions.vm_adjust import VM_ALLOC_TABLE_NAME
from meadowrun.instance_allocation import (
    InstanceRegistrar,
    _InstanceState,
    _TInstanceState,
    allocate_single_job_to_instance,
)
from meadowrun.instance_selection import CloudInstance, ResourcesInternal
from meadowrun.run_job_core import (
    AllocVM,
    CloudProviderType,
    GridJobCloudInterface,
    JobCompletion,
    SshHost,
    WaitOption,
)

if TYPE_CHECKING:
    from types import TracebackType

    import asyncssh
    from meadowrun.abstract_storage_bucket import AbstractStorageBucket
    from meadowrun.meadowrun_pb2 import Job
    from meadowrun.run_job_core import TaskProcessState, WorkerProcessState
    from meadowrun.run_job_local import TaskWorkerServer, WorkerMonitor
    from typing_extensions import Literal


_T = TypeVar("_T")
_U = TypeVar("_U")


@dataclasses.dataclass
class AzureVMInstanceState(_InstanceState):
    # See https://microsoft.github.io/AzureTipsAndTricks/blog/tip88.html
    etag: str


class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
    def __init__(
        self, table_location: str, on_table_missing: Literal["create", "raise"]
    ):
        # this is just the location where the table backing this InstanceRegistrar
        # lives. We can register VMs from any location
        self._table_location = table_location
        self._on_table_missing = on_table_missing

        self._storage_account: Optional[StorageAccount] = None

    async def __aenter__(self) -> AzureInstanceRegistrar:
        self._storage_account = await ensure_table(
            VM_ALLOC_TABLE_NAME, self._table_location, self._on_table_missing
        )
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    def get_region_name(self) -> str:
        return self._table_location

    async def register_instance(
        self,
        public_address: str,
        name: str,
        resources_available: ResourcesInternal,
        running_jobs: List[Tuple[str, ResourcesInternal]],
    ) -> None:
        if self._storage_account is None:
            raise ValueError(
                "Tried to use AzureInstanceRegistrar without calling __aenter__"
            )

        now = datetime.datetime.utcnow().isoformat()
        try:
            # See EC2InstanceRegistrar.register_instance for more details on the schema
            # https://docs.microsoft.com/en-us/rest/api/storageservices/insert-entity
            await azure_table_api(
                "POST",  # inserts rather than replace/update
                self._storage_account,
                VM_ALLOC_TABLE_NAME,
                json_content={
                    "PartitionKey": SINGLE_PARTITION_KEY,
                    "RowKey": name,
                    PUBLIC_ADDRESS: public_address,
                    RESOURCES_AVAILABLE: json.dumps(resources_available.consumable),
                    NON_CONSUMABLE_RESOURCES: json.dumps(
                        resources_available.non_consumable
                    ),
                    RUNNING_JOBS: json.dumps(
                        {
                            job_id: {
                                RESOURCES_ALLOCATED: allocated_resources.consumable,
                                ALLOCATED_TIME: now,
                            }
                            for job_id, allocated_resources in running_jobs
                        }
                    ),
                    PREVENT_FURTHER_ALLOCATION: False,
                    LAST_UPDATE_TIME: now,
                },
            )
        except ResourceExistsError:
            # It's possible that an existing EC2 instance crashed unexpectedly, the
            # coordinator record hasn't been deleted yet, and a new instance was created
            # that has the same address
            raise ValueError(
                f"Tried to register a VM {public_address} but it already exists,"
                " this should never happen!"
            )

    async def get_registered_instances(self) -> List[AzureVMInstanceState]:
        """
        For the AzureInstanceRegistrar, we always populate all the fields on
        AzureVMInstanceState because we will need them in allocate_jobs_to_instance and
        deallocate_job_from_instance
        """
        if self._storage_account is None:
            raise ValueError(
                "Tried to use AzureInstanceRegistrar without calling __aenter__"
            )

        # https://docs.microsoft.com/en-us/rest/api/storageservices/query-entities
        return [
            AzureVMInstanceState(
                item[PUBLIC_ADDRESS],
                item["RowKey"],
                ResourcesInternal(
                    json.loads(item[RESOURCES_AVAILABLE]),
                    json.loads(item[NON_CONSUMABLE_RESOURCES]),
                ),
                json.loads(item[RUNNING_JOBS]),
                item[PREVENT_FURTHER_ALLOCATION],
                item["odata.etag"],
            )
            async for page in azure_table_api_paged(
                "GET",
                self._storage_account,
                VM_ALLOC_TABLE_NAME,
                query_parameters={
                    "$select": ",".join(
                        [
                            "RowKey",
                            RESOURCES_AVAILABLE,
                            NON_CONSUMABLE_RESOURCES,
                            RUNNING_JOBS,
                            PUBLIC_ADDRESS,
                            PREVENT_FURTHER_ALLOCATION,
                        ]
                    ),
                },
            )
            for item in page["value"]
        ]

    async def get_registered_instance(self, name: str) -> AzureVMInstanceState:
        """
        For the AzureInstanceRegistrar, we always populate all the fields on
        AzureVMInstanceState because we will need them in allocate_jobs_to_instance and
        deallocate_job_from_instance
        """
        if self._storage_account is None:
            raise ValueError(
                "Tried to use AzureInstanceRegistrar without calling __aenter__"
            )

        try:
            item = await azure_table_api(
                "GET",
                self._storage_account,
                table_key_url(VM_ALLOC_TABLE_NAME, SINGLE_PARTITION_KEY, name),
                query_parameters={
                    "$select": ",".join(
                        [
                            "RowKey",
                            RESOURCES_AVAILABLE,
                            NON_CONSUMABLE_RESOURCES,
                            RUNNING_JOBS,
                            PUBLIC_ADDRESS,
                            PREVENT_FURTHER_ALLOCATION,
                        ]
                    )
                },
            )
        except ResourceNotFoundError:
            raise ValueError(f"VM {name} was not found")

        return AzureVMInstanceState(
            item[PUBLIC_ADDRESS],
            item["RowKey"],
            ResourcesInternal(
                json.loads(item[RESOURCES_AVAILABLE]),
                json.loads(item[NON_CONSUMABLE_RESOURCES]),
            ),
            json.loads(item[RUNNING_JOBS]),
            item[PREVENT_FURTHER_ALLOCATION],
            item["odata.etag"],
        )

    async def allocate_jobs_to_instance(
        self,
        instance: AzureVMInstanceState,
        resources_allocated_per_job: ResourcesInternal,
        new_job_ids: List[str],
    ) -> bool:
        if self._storage_account is None:
            raise ValueError(
                "Tried to use AzureInstanceRegistrar without calling __aenter__"
            )

        now = datetime.datetime.utcnow().isoformat()

        if len(new_job_ids) == 0:
            raise ValueError("Must provide at least one new_job_ids")

        new_running_jobs = instance.get_running_jobs().copy()
        for job_id in new_job_ids:
            if job_id in instance.get_running_jobs():
                # TODO this should probably be an exception?
                return False

            new_running_jobs[job_id] = {
                RESOURCES_ALLOCATED: resources_allocated_per_job.consumable,
                ALLOCATED_TIME: now,
            }

        new_resources_available = instance.get_available_resources().subtract(
            resources_allocated_per_job.multiply(len(new_job_ids))
        )
        if new_resources_available is None:
            return False  # this shouldn't really happen

        try:
            # https://docs.microsoft.com/en-us/rest/api/storageservices/update-entity2
            await azure_table_api(
                "PUT",  # replace rather than merge
                self._storage_account,
                table_key_url(VM_ALLOC_TABLE_NAME, SINGLE_PARTITION_KEY, instance.name),
                json_content={
                    PUBLIC_ADDRESS: instance.public_address,
                    RESOURCES_AVAILABLE: json.dumps(new_resources_available.consumable),
                    NON_CONSUMABLE_RESOURCES: json.dumps(
                        new_resources_available.non_consumable
                    ),
                    RUNNING_JOBS: json.dumps(new_running_jobs),
                    LAST_UPDATE_TIME: now,
                },
                additional_headers={"If-Match": instance.etag},
            )
            return True
        except ResourceModifiedError:
            # this is how the API indicates that the etag does not match, i.e. the
            # optimistic concurrency check failed
            return False

    async def deallocate_job_from_instance(
        self, instance: AzureVMInstanceState, job_id: str
    ) -> bool:
        if self._storage_account is None:
            raise ValueError(
                "Tried to use AzureInstanceRegistrar without calling __aenter__"
            )

        if job_id not in instance.get_running_jobs():
            return False

        job = instance.get_running_jobs()[job_id]
        new_resources_available = instance.get_available_resources().add(
            ResourcesInternal(job[RESOURCES_ALLOCATED], {})
        )
        new_running_jobs = instance.get_running_jobs().copy()
        del new_running_jobs[job_id]
        now = datetime.datetime.utcnow().isoformat()

        try:
            # https://docs.microsoft.com/en-us/rest/api/storageservices/update-entity2
            await azure_table_api(
                "PUT",
                self._storage_account,
                table_key_url(VM_ALLOC_TABLE_NAME, SINGLE_PARTITION_KEY, instance.name),
                json_content={
                    PUBLIC_ADDRESS: instance.public_address,
                    RESOURCES_AVAILABLE: json.dumps(new_resources_available.consumable),
                    NON_CONSUMABLE_RESOURCES: json.dumps(
                        new_resources_available.non_consumable
                    ),
                    RUNNING_JOBS: json.dumps(new_running_jobs),
                    LAST_UPDATE_TIME: now,
                },
                additional_headers={"If-Match": instance.etag},
            )
            return True
        except ResourceModifiedError:
            # this is how the API indicates that the etag does not match, i.e. the
            # optimistic concurrency check failed
            return False

    async def set_prevent_further_allocation(self, name: str, value: bool) -> bool:
        if self._storage_account is None:
            raise ValueError(
                "Tried to use AzureInstanceRegistrar without calling __aenter__"
            )

        instance = await self.get_registered_instance(name)

        now = datetime.datetime.utcnow().isoformat()

        try:
            # https://docs.microsoft.com/en-us/rest/api/storageservices/update-entity2
            await azure_table_api(
                "PUT",  # replace rather than merge
                self._storage_account,
                table_key_url(VM_ALLOC_TABLE_NAME, SINGLE_PARTITION_KEY, instance.name),
                json_content={
                    PUBLIC_ADDRESS: instance.public_address,
                    RESOURCES_AVAILABLE: json.dumps(
                        instance.get_available_resources().consumable
                    ),
                    NON_CONSUMABLE_RESOURCES: json.dumps(
                        instance.get_available_resources().non_consumable
                    ),
                    RUNNING_JOBS: json.dumps(instance.running_jobs),
                    PREVENT_FURTHER_ALLOCATION: value,
                    LAST_UPDATE_TIME: now,
                },
                additional_headers={"If-Match": instance.etag},
            )
            return True
        except ResourceModifiedError:
            # this is how the API indicates that the etag does not match, i.e. the
            # optimistic concurrency check failed
            return False

    async def launch_instances(
        self,
        resources_required_per_task: ResourcesInternal,
        num_concurrent_tasks: int,
        alloc_cloud_instances: AllocVM,
        abort: Optional[asyncio.Event],
    ) -> Sequence[CloudInstance]:
        if not isinstance(alloc_cloud_instances, AllocAzureVM):
            # TODO do this in the type checker somehow
            raise ValueError(
                "Programming error: AzureInstanceRegistrar can only be used with "
                "AllocAzureVM"
            )

        region_name = alloc_cloud_instances._get_location()
        return await meadowrun.azure_integration.azure_vms.launch_vms(
            resources_required_per_task,
            num_concurrent_tasks,
            (await ensure_meadowrun_key_pair(region_name))[1],
            region_name,
        )

    async def authorize_current_ip(self, alloc_cloud_instances: AllocVM) -> None:
        # TODO currently Azure instances are open to everyone
        pass

    async def open_ports(
        self,
        ports: Optional[Sequence[str]],
        allocated_existing_instances: Iterable[_TInstanceState],
        allocated_new_instances: Iterable[CloudInstance],
    ) -> None:
        if ports:
            raise NotImplementedError(
                "Opening ports on Azure is not implemented, please comment on "
                "https://github.com/meadowdata/meadowrun/issues/126"
            )


@dataclasses.dataclass
class AllocAzureVM(AllocVM):
    """
    Specifies that the job should be run on a dynamically allocated Azure VM. Any
    existing Meadowrun-managed VMs will be reused if available. If none are available,
    Meadowrun will launch the cheapest VM type that meets the resource requirements for
    a job.

    `resources_required` must be provided with the AllocAzureVM Host.

    Attributes:
        location: Specifies the location for the Azure VM, e.g. "eastus". None will use
            the default location.
    """

    location: Optional[str] = None

    def get_cloud_provider(self) -> CloudProviderType:
        return "AzureVM"

    async def set_defaults(self) -> None:
        self.location = get_default_location()

    def _get_location(self) -> str:
        # Small wrapper around region_name that throws if it is None. Should only be
        # used internally.
        if self.location is None:
            raise ValueError(
                "Programming error: location is None but it should have been set by "
                "set_defaults earlier"
            )
        return self.location

    def get_runtime_resources(self) -> ResourcesInternal:
        return ResourcesInternal({}, {})

    async def run_job(
        self,
        resources_required: Optional[ResourcesInternal],
        job: Job,
        wait_for_result: WaitOption,
    ) -> JobCompletion[Any]:
        if resources_required is None:
            raise ValueError(
                "Resources.logical_cpu and memory_gb must be specified for "
                "AllocAzureVM"
            )

        return await run_job_azure_vm_instance_registrar(
            job, resources_required, self, wait_for_result
        )

    def _create_grid_job_cloud_interface(self) -> GridJobCloudInterface:
        return AzureVMGridJobInterface(self)

    async def get_storage_bucket(self) -> AbstractStorageBucket:
        return await get_azure_blob_container(self._get_location())


async def run_job_azure_vm_instance_registrar(
    job: Job,
    resources_required: ResourcesInternal,
    alloc_azure_vm: AllocAzureVM,
    wait_for_result: WaitOption,
) -> JobCompletion[Any]:
    location = alloc_azure_vm._get_location()

    pkey, public_key = await ensure_meadowrun_key_pair(location)

    async with AzureInstanceRegistrar(location, "create") as instance_registrar:
        if job.ports:
            raise NotImplementedError(
                "Opening ports on Azure is not implemented, please comment on "
                "https://github.com/meadowdata/meadowrun/issues/126"
            )
        host, job_id = await allocate_single_job_to_instance(
            instance_registrar,
            resources_required,
            alloc_azure_vm,
            job.ports,
        )

    # Kind of weird that we're changing the job_id here, but okay as long as job_id
    # remains mostly an internal concept
    job.job_id = job_id

    return await SshHost(host, "meadowrunuser", pkey, ("AzureVM", location)).run_job(
        resources_required, job, wait_for_result
    )


class AzureVMGridJobInterface(GridJobCloudInterface, Generic[_T, _U]):
    """
    This is a relatively thin wrapper around some functionality in grid_tasks_queue.
    This class should be in grid_tasks_queue, but it's here because of circular import
    issues.
    """

    def create_queue(self) -> int:
        raise NotImplementedError(
            "Retrying with more resources is not supported on Azure yet"
        )

    def __init__(self, alloc_cloud_instance: AllocAzureVM):
        self._cloud_provider = alloc_cloud_instance.get_cloud_provider()
        self._location = alloc_cloud_instance._get_location()

        self._ssh_private_key: Optional[
            asyncio.Task[Tuple[asyncssh.SSHKey, str]]
        ] = None

        self._request_result_queues: Optional[asyncio.Task[Tuple[Queue, Queue]]] = None

        self._tasks: Optional[Sequence[_T]] = None

        # this id is just used for creating the job's queues. It has no relationship to
        # any Job.job_ids
        self._job_id = str(uuid.uuid4())

    def create_instance_registrar(self) -> InstanceRegistrar:
        return AzureInstanceRegistrar(self._location, "create")

    async def setup_and_add_tasks(self, tasks: Sequence[_T]) -> None:
        print(f"The current run_map's id is {self._job_id}")
        self._ssh_private_key = asyncio.create_task(
            ensure_meadowrun_key_pair(self._location)
        )
        self._request_result_queues = asyncio.create_task(
            create_queues_for_job(self._job_id, self._location)
        )
        self._tasks = tasks
        await add_tasks((await self._request_result_queues)[0], tasks)

    async def ssh_host_from_address(self, address: str, instance_name: str) -> SshHost:
        if self._ssh_private_key is None:
            raise ValueError(
                "Must call setup_and_add_tasks before calling ssh_host_from_address"
            )

        return SshHost(
            address,
            "meadowrunuser",
            (await self._ssh_private_key)[0],
            (self._cloud_provider, self._location),
            instance_name,
        )

    async def shutdown_workers(self, num_workers: int, queue_index: int) -> None:
        if queue_index != 0:
            raise NotImplementedError(
                "Retrying with more resources is not implemented for Azure yet"
            )

        if self._request_result_queues is None:
            raise ValueError(
                "Must call setup_and_add_tasks before calling shutdown_workers"
            )

        await add_worker_shutdown_message(
            (await self._request_result_queues)[0], num_workers
        )

    async def get_worker_function(
        self, queue_index: int
    ) -> Callable[
        [str, str, TaskWorkerServer, WorkerMonitor],
        Coroutine[Any, Any, None],
    ]:
        if self._request_result_queues is None:
            raise ValueError(
                "Must call setup_and_add_tasks before calling get_worker_function"
            )
        if queue_index != 0:
            raise NotImplementedError(
                "Retrying with more resources is not implemented for Azure yet"
            )

        request_queue, result_queue = await self._request_result_queues

        return functools.partial(
            worker_function,
            request_queue,
            result_queue,
        )

    async def receive_task_results(
        self, *, stop_receiving: asyncio.Event, workers_done: asyncio.Event
    ) -> AsyncIterable[Tuple[List[TaskProcessState], List[WorkerProcessState]]]:
        if self._request_result_queues is None:
            raise ValueError(
                "Must call setup_and_add_tasks before calling receive_task_results"
            )

        return get_results_unordered(
            (await self._request_result_queues)[1],
            self._location,
            stop_receiving,
            workers_done,
        )

    async def retry_task(
        self, task_id: int, attempts_so_far: int, queue_index: int
    ) -> None:
        if self._tasks is None or self._request_result_queues is None:
            raise ValueError("Must call setup_and_add_tasks before calling retry_task")

        await retry_task(
            (await self._request_result_queues)[0],
            task_id,
            attempts_so_far + 1,
            self._tasks[task_id],
        )



from __future__ import annotations

from typing import Tuple


STORAGE_ENV_CACHE_PREFIX = "env_cache/"
STORAGE_CODE_CACHE_PREFIX = "code_cache/"


def storage_key_task_args(job_id: str) -> str:
    return f"inputs/{job_id}.task_args"


def storage_key_ranges(job_id: str) -> str:
    return f"inputs/{job_id}.ranges"


def storage_key_job_to_run(job_id: str) -> str:
    return f"inputs/{job_id}.job_to_run"


def storage_prefix_outputs(job_id: str) -> str:
    return f"outputs/{job_id}/"


def storage_key_task_result(job_id: str, task_id: int, attempt: int) -> str:
    # A million tasks and 1000 attempts should be enough for everybody. Formatting the
    # task is important because when we task download results from S3, we use the
    # StartFrom argument to S3's ListObjects to exclude most tasks we've already
    # downloaded.
    return (
        f"{storage_prefix_outputs(job_id)}{task_id:06d}.{attempt:03d}"
        f"{STORAGE_KEY_TASK_RESULT_SUFFIX}"
    )


STORAGE_KEY_TASK_RESULT_SUFFIX = ".taskresult"


def parse_storage_key_task_result(key: str, results_prefix: str) -> Tuple[int, int]:
    """Returns task_id, attempt based on the task result key"""
    [task_id, attempt, _] = key.replace(results_prefix, "").split(".")
    return int(task_id), int(attempt)


STORAGE_KEY_PROCESS_STATE_SUFFIX = ".process_state"


def storage_key_process_state(job_id: str, worker_index: str) -> str:
    return (
        f"{storage_prefix_outputs(job_id)}{worker_index}"
        f"{STORAGE_KEY_PROCESS_STATE_SUFFIX}"
    )


def parse_storage_key_process_state(key: str, results_prefix: str) -> str:
    return key.replace(results_prefix, "").replace(STORAGE_KEY_PROCESS_STATE_SUFFIX, "")



from __future__ import annotations

import argparse
import asyncio
import asyncio.subprocess
import logging
from typing import TYPE_CHECKING

from meadowrun.aws_integration.aws_core import (
    _get_default_region_name,
    _get_ec2_metadata,
)
from meadowrun.aws_integration.ec2_instance_allocation import EC2InstanceRegistrar
from meadowrun.azure_integration.azure_instance_allocation import AzureInstanceRegistrar
from meadowrun.azure_integration.azure_meadowrun_core import (
    get_current_vm_name,
    get_default_location,
    get_scheduled_events_on_vm,
)

if TYPE_CHECKING:
    from meadowrun.instance_allocation import InstanceRegistrar
    from meadowrun.run_job_core import CloudProviderType
from meadowrun.run_job_core import CloudProvider


async def async_main(cloud: CloudProviderType, cloud_region_name: str) -> None:
    """
    Checks to see if we're running on a spot instance that's scheduled for
    eviction (aka interruption). If so, updates the InstanceRegistrar to prevent new
    jobs from being scheduled on this instance
    """
    if cloud == "EC2":
        if (await _get_ec2_metadata("spot/instance-action")) is None:
            # this means we're not a spot instance that's being interrupted right now
            return

        instance_name = await _get_ec2_metadata("instance-id")
        if cloud_region_name == "default":
            cloud_region_name = await _get_default_region_name()
        instance_registrar: InstanceRegistrar = EC2InstanceRegistrar(
            cloud_region_name, "raise"
        )
    elif cloud == "AzureVM":
        scheduled_events = await get_scheduled_events_on_vm()
        if scheduled_events is None or (
            not any(
                event.get("EventType") == "Preempt"
                for event in scheduled_events["Events"]
            )
        ):
            return

        instance_name = await get_current_vm_name()
        if cloud_region_name == "default":
            cloud_region_name = get_default_location()
        instance_registrar = AzureInstanceRegistrar(cloud_region_name, "raise")
    else:
        raise ValueError(f"Unexpected value for cloud_provider: {cloud}")

    async with instance_registrar:
        if not instance_name:
            raise ValueError(
                "Cannot register spot instance eviction because we can't get the "
                f"name/id of the current {cloud} instance (maybe we're not "
                f"running on a {cloud} instance?)"
            )

        print(
            "We are on a spot instance that is being terminated, we will stop "
            "allocating new jobs to this instance"
        )
        while not await instance_registrar.set_prevent_further_allocation(
            instance_name, True
        ):
            # Just keep trying to set set_prevent_further_allocation if it fails. Our
            # instance will get shutdown soon anyways
            await asyncio.sleep(1)


def main(cloud: CloudProviderType, cloud_region_name: str) -> None:
    asyncio.run(async_main(cloud, cloud_region_name))


def command_line_main() -> None:
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("--cloud", required=True, choices=CloudProvider)
    parser.add_argument("--cloud-region-name", required=True)
    args = parser.parse_args()

    main(args.cloud, args.cloud_region_name)


if __name__ == "__main__":
    command_line_main()



from __future__ import annotations

import abc
import dataclasses
import os
import pathlib
import platform
import sys
import tempfile
import urllib.parse
from typing import (
    Optional,
    Union,
    Sequence,
    Mapping,
    Dict,
    List,
    Iterable,
    cast,
    TYPE_CHECKING,
)

from meadowrun.deployment.local_code import zip_local_code
from meadowrun.azure_integration.azure_ssh_keys import get_meadowrun_vault_name
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    get_subscription_id_sync,
)
from meadowrun.deployment.conda import try_get_current_conda_env, env_export

from meadowrun.credentials import CredentialsService, CredentialsSourceForService

if TYPE_CHECKING:
    from meadowrun.deployment_internal_types import (
        InterpreterDeployment,
        VersionedInterpreterDeployment,
        CodeDeployment,
        VersionedCodeDeployment,
    )
from meadowrun.docker_controller import get_registry_domain
from meadowrun.meadowrun_pb2 import (
    AwsSecretProto,
    AzureSecretProto,
    CodeZipFile,
    ContainerAtDigest,
    ContainerAtTag,
    Credentials,
    EnvironmentSpec,
    EnvironmentSpecInCode,
    EnvironmentType,
    GitRepoBranch,
    GitRepoCommit,
    KubernetesSecretProto,
    ServerAvailableContainer,
    ServerAvailableFolder,
    ServerAvailableInterpreter,
)
from meadowrun.deployment.pip import pip_freeze_exclude_editable


class Secret(abc.ABC):
    """
    An abstract class for specifying a secret, e.g. a username/password or an SSH key
    """

    @abc.abstractmethod
    def _to_credentials_source(
        self,
        service: CredentialsService,
        service_url: str,
        credentials_type: Credentials.Type.ValueType,
    ) -> CredentialsSourceForService:
        pass


@dataclasses.dataclass(frozen=True)
class AwsSecret(Secret):
    """
    An AWS secret

    Attributes:
        secret_name: The name of the secret (also sometimes called the id)
    """

    secret_name: str

    def _to_credentials_source(
        self,
        service: CredentialsService,
        service_url: str,
        credentials_type: Credentials.Type.ValueType,
    ) -> CredentialsSourceForService:
        return CredentialsSourceForService(
            service=service,
            service_url=service_url,
            source=AwsSecretProto(
                credentials_type=credentials_type,
                secret_name=self.secret_name,
            ),
        )


@dataclasses.dataclass(frozen=True)
class AzureSecret(Secret):
    """
    An Azure secret

    Attributes:
        secret_name: The name of the secret
        vault_name: The name of the Key Vault that the secret is in. Defaults to None,
            which implies the Meadowrun-managed Key Vault (mr<last 22 characters of your
            subscription id>)
    """

    secret_name: str
    vault_name: Optional[str] = None

    def _to_credentials_source(
        self,
        service: CredentialsService,
        service_url: str,
        credentials_type: Credentials.Type.ValueType,
    ) -> CredentialsSourceForService:
        if self.vault_name is None:
            vault_name = get_meadowrun_vault_name(get_subscription_id_sync())
        else:
            vault_name = self.vault_name

        return CredentialsSourceForService(
            service=service,
            service_url=service_url,
            source=AzureSecretProto(
                credentials_type=credentials_type,
                secret_name=self.secret_name,
                vault_name=vault_name,
            ),
        )


@dataclasses.dataclass(frozen=True)
class KubernetesSecret(Secret):
    """
    A Kubernetes secret

    Attributes:
        secret_name: The name of the secret
    """

    secret_name: str

    def _to_credentials_source(
        self,
        service: CredentialsService,
        service_url: str,
        credentials_type: Credentials.Type.ValueType,
    ) -> CredentialsSourceForService:
        return CredentialsSourceForService(
            service=service,
            service_url=service_url,
            source=KubernetesSecretProto(
                credentials_type=credentials_type, secret_name=self.secret_name
            ),
        )


@dataclasses.dataclass(frozen=True)
class PreinstalledInterpreter:
    """
    Represents an interpreter that has been pre-installed on the remote machine. This is
    useful if you're using a [custom AMI](../../how_to/custom_ami).

    Attributes:
        path_to_interpreter: The path to the python executable, e.g.
            `/var/myenv/bin/python`
    """

    path_to_interpreter: str


class ContainerInterpreterBase(abc.ABC):
    """An abstract base class for specifying a container as a python interpreter"""

    @abc.abstractmethod
    def get_interpreter_spec(
        self,
    ) -> Union[InterpreterDeployment, VersionedInterpreterDeployment]:
        pass

    @abc.abstractmethod
    def _get_repository_name(self) -> str:
        pass

    @abc.abstractmethod
    def _get_username_password_secret(self) -> Optional[Secret]:
        pass


@dataclasses.dataclass(frozen=True)
class ContainerInterpreter(ContainerInterpreterBase):
    """
    Specifies a container image. The container image must be configured so that running
    `docker run -it repository_name:tag python` runs the right python interpreter.

    Attributes:
        repository_name: E.g. `python`, or `<my_azure_registry>.azurecr.io/foo`
        tag: E.g. `3.9-slim-bullseye` or `latest` (the default)
        username_password_secret: An AWS or Azure secret that has a username and
            password for connecting to the container registry (as specified or implied
            in image_name). Only needed if the image/container registry is private.
        always_use_local: If this is True, only looks for the image on the EC2 instance
            and does not try to download the image from a container registry. This will
            only work if you've preloaded the image into the AMI via [Use a custom AMI
            on AWS](../../how_to/custom_ami)
    """

    repository_name: str
    tag: str = "latest"
    username_password_secret: Optional[Secret] = None
    always_use_local: bool = False

    def get_interpreter_spec(
        self,
    ) -> Union[InterpreterDeployment, VersionedInterpreterDeployment]:
        if self.always_use_local:
            return ServerAvailableContainer(
                image_name=f"{self.repository_name}:{self.tag}"
            )
        else:
            return ContainerAtTag(repository=self.repository_name, tag=self.tag)

    def _get_repository_name(self) -> str:
        return self.repository_name

    def _get_username_password_secret(self) -> Optional[Secret]:
        return self.username_password_secret


@dataclasses.dataclass(frozen=True)
class ContainerAtDigestInterpreter(ContainerInterpreterBase):
    """
    Like [ContainerInterpreter][meadowrun.ContainerInterpreter] but specifies a digest
    instead of a tag. Running `docker run -it repository_name@digest python` must run
    the right python interpreter.

    Attributes:
        repository_name: E.g. `python`, or `<my_azure_registry>.azurecr.io/foo`
        digest: E.g. `sha256:97725c608...`
        username_password_secret: An AWS or Azure secret that has a username and
            password for connecting to the container registry (as specified or implied
            in image_name). Only needed if the image/container registry is private.
        always_use_local: If this is True, only looks for the image on the EC2 instance
            and does not try to download the image from a container registry. This will
            only work if you've preloaded the image into the AMI via [Use a custom AMI
            on AWS](../../how_to/custom_ami)
    """

    repository_name: str
    digest: str
    username_password_secret: Optional[Secret] = None
    always_use_local: bool = False

    def get_interpreter_spec(
        self,
    ) -> Union[InterpreterDeployment, VersionedInterpreterDeployment]:
        if self.always_use_local:
            return ServerAvailableContainer(
                image_name=f"{self.repository_name}@{self.digest}"
            )
        else:
            return ContainerAtDigest(
                repository=self.repository_name, digest=self.digest
            )

    def _get_repository_name(self) -> str:
        return self.repository_name

    def _get_username_password_secret(self) -> Optional[Secret]:
        return self.username_password_secret


def _additional_software_to_dict(
    additional_software: Union[Sequence[str], str, None],
) -> Mapping[str, str]:
    if additional_software is None:
        return {}

    if isinstance(additional_software, str):
        return {additional_software: ""}

    return {package: "" for package in additional_software}


class InterpreterSpecFile(abc.ABC):
    """
    An abstract base class for specifying a Conda environment.yml file or a pip
    requirements.txt file
    """

    pass


@dataclasses.dataclass(frozen=True)
class CondaEnvironmentYmlFile(InterpreterSpecFile):
    """
    Specifies an environment.yml file generated by conda env export.

    Attributes:
        path_to_yml_file: In the context of mirror_local, this is the path to a file on
            the local disk. In the context of git_repo, this is a path to a file in the
            git repo
        additional_software: apt packages that need to be installed to make the
            environment work
    """

    path_to_yml_file: str
    additional_software: Union[Sequence[str], str, None] = None


@dataclasses.dataclass(frozen=True)
class PipRequirementsFile(InterpreterSpecFile):
    """
    Specifies a requirements.txt file generated by pip freeze.

    Attributes:
        path_to_requirements_file: In the context of mirror_local, this is the path to a
            file on the local disk. In the context of git_repo, this is a path to a file
            in the git repo
        python_version: A python version like "3.9" or "3.9.5". The version must be
            available on docker: https://hub.docker.com/_/python as
            python-<python_version>-slim-bullseye.
        additional_software: apt packages that need to be installed to make the
            environment work
    """

    path_to_requirements_file: str
    python_version: str
    additional_software: Union[Sequence[str], str, None] = None


@dataclasses.dataclass(frozen=True)
class PoetryProjectPath(InterpreterSpecFile):
    """
    Specifies a poetry project

    Attributes:
        path_to_project: In the context of mirror_local, this is the path to a folder on
            the local disk that contains pyproject.toml and poetry.lock. In the context
            of git_repo, this is a path to a folder in the git repo (use `""` to
            indicate that pyproject.toml and poetry.lock are at the root of the repo).
        python_version: A python version like "3.9" or "3.9.5". The version must be
            available on docker: https://hub.docker.com/_/python as
            python-<python_version>-slim-bullseye. This python version must be
            compatible with the requirements in pyproject.toml
        additional_software: apt packages that need to be installed to make the
            environment work
    """

    path_to_project: str
    python_version: str
    additional_software: Union[Sequence[str], str, None] = None


class LocalInterpreter(abc.ABC):
    """
    An abstract base class for specifying a python interpreter on the local machine
    """

    @abc.abstractmethod
    async def get_interpreter_spec(self) -> InterpreterDeployment:
        pass


@dataclasses.dataclass(frozen=True)
class LocalCurrentInterpreter(LocalInterpreter):
    """
    Specifies the current python interpreter.

    Attributes:
        additional_software: apt packages that need to be installed to make the
            environment work
    """

    additional_software: Union[Sequence[str], str, None] = None

    async def get_interpreter_spec(self) -> InterpreterDeployment:
        # first, check if we are in a conda environment
        conda_env_spec = await try_get_current_conda_env()
        if conda_env_spec is not None:
            if platform.system() != "Linux":
                raise ValueError(
                    "mirror_local from a Conda environment only works from Linux "
                    "because conda environments are not cross-platform."
                )
            print("Mirroring current conda environment")
            return EnvironmentSpec(
                environment_type=EnvironmentType.CONDA,
                spec=conda_env_spec,
                additional_software=_additional_software_to_dict(
                    self.additional_software
                ),
            )

        # next, check if we're in a poetry environment. Unfortunately it doesn't seem
        # like there's a way to detect that we're in a poetry environment unless the
        # current working directory or a parent contains the pyproject.toml and
        # poetry.lock files.
        cwd = pathlib.Path.cwd()
        candidates = [cwd]
        candidates.extend(cwd.parents)
        is_poetry_env = False
        for path in candidates:
            pyproject_file = path / "pyproject.toml"
            poetry_lock_file = path / "poetry.lock"

            if pyproject_file.exists() and poetry_lock_file.exists():
                is_poetry_env = True
                break

        if is_poetry_env:
            with open(pyproject_file, encoding="utf-8") as project_file:
                project_file_contents = project_file.read()
            with open(poetry_lock_file, encoding="utf-8") as lock_file:
                lock_file_contents = lock_file.read()
            print("Mirroring current poetry environment")
            return EnvironmentSpec(
                environment_type=EnvironmentType.POETRY,
                spec=project_file_contents,
                spec_lock=lock_file_contents,
                python_version=f"{sys.version_info.major}.{sys.version_info.minor}",
                additional_software=_additional_software_to_dict(
                    self.additional_software
                ),
            )

        # last resort, assume this is a pip-based environment
        print("Mirroring current pip environment")
        return EnvironmentSpec(
            environment_type=EnvironmentType.PIP,
            spec=await pip_freeze_exclude_editable(),
            python_version=f"{sys.version_info.major}.{sys.version_info.minor}",
            additional_software=_additional_software_to_dict(self.additional_software),
        )


@dataclasses.dataclass(frozen=True)
class LocalCondaInterpreter(LocalInterpreter):
    """
    Specifies a locally installed conda environment

    Attributes:
        environment_name_or_path: Either the name of a conda environment (e.g.
            `my_env_name`) or the full path to the folder of a conda environment (e.g.
            `/home/user/miniconda3/envs/my_env_name`). Will be passed to conda env
            export
        additional_software: apt packages that need to be installed to make the
            environment work
    """

    environment_name_or_path: str
    additional_software: Union[Sequence[str], str, None] = None

    async def get_interpreter_spec(self) -> InterpreterDeployment:
        if platform.system() != "Linux":
            raise ValueError(
                "mirror_local from a Conda environment only works from Linux because "
                "conda environments are not cross-platform."
            )
        return EnvironmentSpec(
            environment_type=EnvironmentType.CONDA,
            spec=await env_export(self.environment_name_or_path),
            additional_software=_additional_software_to_dict(self.additional_software),
        )


@dataclasses.dataclass(frozen=True)
class LocalPipInterpreter(LocalInterpreter):
    """
    Specifies a locally available interpreter. It can be a "regular" install of a python
    interpreter, a virtualenv, or anything based on pip.

    Attributes:
        path_to_interpreter: The path to the python executable. E.g.
            `/home/user/my_virtual_env/bin/python`
        python_version: A python version like "3.9" or "3.9.5". The version must be
            available on docker: https://hub.docker.com/_/python as
            python-<python_version>-slim-bullseye.
        additional_software: apt packages that need to be installed to make the
            environment work
    """

    path_to_interpreter: str
    python_version: str
    additional_software: Union[Sequence[str], str, None] = None

    async def get_interpreter_spec(self) -> InterpreterDeployment:
        return EnvironmentSpec(
            environment_type=EnvironmentType.PIP,
            # TODO this won't work if the specified environment has editable installs
            spec=await pip_freeze_exclude_editable(self.path_to_interpreter),
            python_version=self.python_version,
            additional_software=_additional_software_to_dict(self.additional_software),
        )


@dataclasses.dataclass(frozen=True)
class Deployment:
    interpreter: Union[
        InterpreterDeployment, VersionedInterpreterDeployment, None
    ] = None
    code: Union[CodeDeployment, VersionedCodeDeployment, None] = None
    environment_variables: Optional[Dict[str, str]] = None
    credentials_sources: List[CredentialsSourceForService] = dataclasses.field(
        default_factory=list
    )

    @classmethod
    async def mirror_local(
        cls,
        include_sys_path: bool = True,
        additional_python_paths: Union[Iterable[str], str] = tuple(),
        interpreter: Union[
            LocalInterpreter, InterpreterSpecFile, ContainerInterpreterBase, None
        ] = None,
        working_directory_globs: Union[str, Iterable[str], None] = None,
        environment_variables: Optional[Dict[str, str]] = None,
    ) -> Deployment:
        """A deployment that mirrors the local environment and code.

        Args:
            include_sys_path: if True, find python code in the paths in sys.path
                (effectively "your local code"), copies them to the remote machines, and
                add them to sys.path there. Ignores any installed packages.
            additional_python_paths: local code paths to copy and add to python path on
                the remote machine.
            interpreter: Specifies the environment/interpreter to use. Defaults to None
                which will detect the currently activated env. Alternatively, you can
                explicitly specify a locally installed python environment with
                [LocalCondaInterpreter][meadowrun.LocalCondaInterpreter],
                [LocalPipInterpreter][meadowrun.LocalPipInterpreter],
                [CondaEnvironmentYmlFile][meadowrun.CondaEnvironmentYmlFile],
                [PipRequirementsFile][meadowrun.PipRequirementsFile],
                [PoetryProjectPath][meadowrun.PoetryProjectPath]
            working_directory_globs: Most of the time, your current working directory
                will be on `sys.path`, and so any *.py files from your current working
                directory will be uploaded by `mirror_local`. However, any other types
                of files will be ignored by default. You can specify files to include
                from your current working directory with this argument. Examples:
                `"foo/bar.txt"` will include the specified file. `"*.txt"` will specify
                txt files in your current directory (but not recursively). `"**/*.txt"`
                will specify all txt files in your current directory recursively (e.g.
                will capture both `1.txt` and `foo/2.txt`). `"foo/**/*.txt"` will
                capture all txt files in the foo directory. If you specify files like
                this, you will be able to access them with relative paths in the remote
                code the same way you reference them locally.
            environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
                variables will be set in the remote environment.

        Returns:
            A `Deployment` object that can be passed to the `run_*` functions.
        """

        credentials_sources = []

        if interpreter is None:
            interpreter = LocalCurrentInterpreter()

        if isinstance(interpreter, CondaEnvironmentYmlFile):
            with open(interpreter.path_to_yml_file, encoding="utf-8") as f:
                interpreter_spec: Union[
                    InterpreterDeployment, VersionedInterpreterDeployment
                ] = EnvironmentSpec(
                    environment_type=EnvironmentType.CONDA,
                    spec=f.read(),
                    additional_software=_additional_software_to_dict(
                        interpreter.additional_software
                    ),
                )
        elif isinstance(interpreter, PipRequirementsFile):
            with open(interpreter.path_to_requirements_file, encoding="utf-8") as f:
                interpreter_spec = EnvironmentSpec(
                    environment_type=EnvironmentType.PIP,
                    spec=f.read(),
                    python_version=interpreter.python_version,
                    additional_software=_additional_software_to_dict(
                        interpreter.additional_software
                    ),
                )
        elif isinstance(interpreter, PoetryProjectPath):
            with open(
                os.path.join(interpreter.path_to_project, "pyproject.toml"),
                "r",
                encoding="utf-8",
            ) as spec_file:
                spec = spec_file.read()
            with open(
                os.path.join(interpreter.path_to_project, "poetry.lock"),
                "r",
                encoding="utf-8",
            ) as spec_lock_file:
                spec_lock = spec_lock_file.read()
            interpreter_spec = EnvironmentSpec(
                environment_type=EnvironmentType.POETRY,
                spec=spec,
                spec_lock=spec_lock,
                python_version=interpreter.python_version,
                additional_software=_additional_software_to_dict(
                    interpreter.additional_software
                ),
            )
        elif isinstance(interpreter, LocalInterpreter):
            interpreter_spec = await interpreter.get_interpreter_spec()
        elif isinstance(interpreter, ContainerInterpreterBase):
            interpreter_spec = interpreter.get_interpreter_spec()
            username_password_secret = interpreter._get_username_password_secret()
            if username_password_secret is not None:
                credentials_sources.append(
                    username_password_secret._to_credentials_source(
                        "DOCKER",
                        get_registry_domain(interpreter._get_repository_name())[0],
                        Credentials.Type.USERNAME_PASSWORD,
                    )
                )
        elif isinstance(interpreter, PreinstalledInterpreter):
            interpreter_spec = ServerAvailableInterpreter(
                interpreter_path=interpreter.path_to_interpreter
            )
        else:
            raise ValueError(f"Unexpected type of interpreter {type(interpreter)}")

        if working_directory_globs is None:
            working_directory_globs = ()
        elif isinstance(working_directory_globs, str):
            working_directory_globs = [working_directory_globs]

        if isinstance(additional_python_paths, str):
            additional_python_paths = [additional_python_paths]

        # annoyingly, this tmp dir now gets deleted in run_local when the file
        # has been uploaded/unpacked depending on the Host implementation
        tmp_dir = tempfile.mkdtemp()
        zip_file_path, zip_python_paths, zip_cwd = zip_local_code(
            tmp_dir,
            include_sys_path,
            additional_python_paths,
            working_directory_globs=working_directory_globs,
        )

        # see comment on _upload_code_zip_file
        url = urllib.parse.urlunparse(("file", "", zip_file_path, "", "", ""))
        code = CodeZipFile(url=url, code_paths=zip_python_paths, cwd_path=zip_cwd)

        return cls(interpreter_spec, code, environment_variables, credentials_sources)

    @classmethod
    def git_repo(
        cls,
        repo_url: str,
        branch: Optional[str] = None,
        commit: Optional[str] = None,
        path_to_source: Optional[str] = None,
        interpreter: Union[InterpreterSpecFile, ContainerInterpreterBase, None] = None,
        environment_variables: Optional[Dict[str, str]] = None,
        ssh_key_secret: Optional[Secret] = None,
    ) -> Deployment:
        """
        A deployment based on a git repo.

        Args:
            repo_url: e.g. `"https://github.com/meadowdata/test_repo"`
            branch: defaults to `"main"` if neither branch nor commit are specified.
            commit: can be provided instead of branch to use a specific commit hash,
                e.g. `"d018b54"`
            path_to_source: e.g. `"src/python"` to use a subdirectory of the repo
            interpreter: specifies either a Conda environment.yml or requirements.txt
                file in the Git repo (relative to the repo, note that this IGNORES
                `path_to_source`) This file will be used to generate the environment to
                run in. See
                [CondaEnvironmentYmlFile][meadowrun.CondaEnvironmentYmlFile],
                [PipRequirementsFile][meadowrun.PipRequirementsFile], and
                [PoetryProjectPath][meadowrun.PoetryProjectPath]
            environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
                variables will be set in the remote environment
            ssh_key_secret: A secret that contains the contents of a private SSH key
                that has read access to `repo_url`, e.g. `AwsSecret("my_ssh_key")`. See
                How to use a private git repo for [AWS](/how_to/private_git_repo_aws)
                or [Azure](/how_to/private_git_repo_azure)

        Returns:
            A `Deployment` object that can be passed to the `run_*` functions.
        """
        if branch and commit:
            raise ValueError("Only one of branch and commit can be specified")

        if not branch and not commit:
            branch = "main"

        if branch:
            code: Union[CodeDeployment, VersionedCodeDeployment, None] = GitRepoBranch(
                repo_url=repo_url,
                branch=branch,
                # the generated protobuf types say that path_to_source must not be None,
                # but this is not actually the case, it handles None fine, which we want
                # to take advantage of
                path_to_source=cast(str, path_to_source),
            )
        else:  # commit
            # guaranteed by prior ifs, this is just for mypy
            assert commit is not None
            code = GitRepoCommit(
                repo_url=repo_url,
                commit=commit,
                # see comment above
                path_to_source=cast(str, path_to_source),
            )

        credentials_sources = []

        if interpreter is None:
            interpreter_spec: Union[
                InterpreterDeployment, VersionedInterpreterDeployment, None
            ] = None
        elif isinstance(interpreter, CondaEnvironmentYmlFile):
            interpreter_spec = EnvironmentSpecInCode(
                environment_type=EnvironmentType.CONDA,
                path_to_spec=interpreter.path_to_yml_file,
                additional_software=_additional_software_to_dict(
                    interpreter.additional_software
                ),
            )
        elif isinstance(interpreter, PipRequirementsFile):
            interpreter_spec = EnvironmentSpecInCode(
                environment_type=EnvironmentType.PIP,
                path_to_spec=interpreter.path_to_requirements_file,
                python_version=interpreter.python_version,
                additional_software=_additional_software_to_dict(
                    interpreter.additional_software
                ),
            )
        elif isinstance(interpreter, PoetryProjectPath):
            interpreter_spec = EnvironmentSpecInCode(
                environment_type=EnvironmentType.POETRY,
                path_to_spec=interpreter.path_to_project,
                python_version=interpreter.python_version,
                additional_software=_additional_software_to_dict(
                    interpreter.additional_software
                ),
            )
        elif isinstance(interpreter, ContainerInterpreterBase):
            interpreter_spec = interpreter.get_interpreter_spec()
            username_password_secret = interpreter._get_username_password_secret()
            if username_password_secret is not None:
                credentials_sources.append(
                    username_password_secret._to_credentials_source(
                        "DOCKER",
                        get_registry_domain(interpreter._get_repository_name())[0],
                        Credentials.Type.USERNAME_PASSWORD,
                    )
                )
        elif isinstance(interpreter, PreinstalledInterpreter):
            interpreter_spec = ServerAvailableInterpreter(
                interpreter_path=interpreter.path_to_interpreter
            )
        else:
            raise ValueError(f"Unexpected type of interpreter: {type(interpreter)}")

        if ssh_key_secret is not None:
            credentials_sources.append(
                ssh_key_secret._to_credentials_source(
                    "GIT", repo_url, Credentials.Type.SSH_KEY
                )
            )

        return cls(interpreter_spec, code, environment_variables, credentials_sources)

    @classmethod
    def container_image(
        cls,
        repository: str,
        tag: str = "latest",
        username_password_secret: Optional[Secret] = None,
        environment_variables: Optional[Dict[str, str]] = None,
    ) -> Deployment:
        """
        A deployment based on a docker container image

        Arguments:
            repository: The name of the docker container image repository, e.g. `python`
                or `quay.io/minio/minio`.
            tag: Combined with repository, will be used like `{repository}:{tag}`.
                Defaults to `latest`
            environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
                variables will be set in the remote environment.
        """
        credentials_sources = []
        if username_password_secret is not None:
            credentials_sources.append(
                username_password_secret._to_credentials_source(
                    "DOCKER",
                    get_registry_domain(repository)[0],
                    Credentials.Type.USERNAME_PASSWORD,
                )
            )
        return cls(
            ContainerAtTag(repository=repository, tag=tag),
            ServerAvailableFolder(),
            environment_variables,
            credentials_sources,
        )

    @classmethod
    def container_image_at_digest(
        cls,
        repository: str,
        digest: str,
        username_password_secret: Optional[Secret] = None,
        environment_variables: Optional[Dict[str, str]] = None,
    ) -> Deployment:
        """
        A deployment based on a docker container image

        Arguments:
            repository: The name of the docker container image repository, e.g. `python`
                or `quay.io/minio/minio`.
            digest: Combined with repository, will be used like `{repository}@{tag}`.
                Defaults to `latest` if digest is not specified.
            environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
                variables will be set in the remote environment.
        """
        credentials_sources = []
        if username_password_secret is not None:
            credentials_sources.append(
                username_password_secret._to_credentials_source(
                    "DOCKER",
                    get_registry_domain(repository)[0],
                    Credentials.Type.USERNAME_PASSWORD,
                )
            )
        return cls(
            ContainerAtDigest(repository=repository, digest=digest),
            ServerAvailableFolder(),
            environment_variables,
            credentials_sources,
        )



from __future__ import annotations

import abc
from typing import Optional, Type, Tuple, List, TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType


class AbstractStorageBucket(abc.ABC):
    """
    A wrapper around S3, Azure Storage, Google Cloud Storage, or other similar object
    storage systems.

    All of these systems implement an S3-compatible API, so we could theoretically get
    by with just S3ClientWrapper which uses boto3 under the hood, but the big difference
    is authentication. For example, Google Cloud Storage "natively" uses OAuth2, where
    as S3/boto3 use an HMAC-based authentication scheme. So if we want to use e.g. a
    service account's identity-based permissions on a GKE node, we need to use the
    Google Cloud API directly, as S3ClientWrapper/boto3 will only work with a
    username/password.
    """

    async def __aenter__(self) -> AbstractStorageBucket:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    @abc.abstractmethod
    def get_cache_key(self) -> str:
        """
        The cache key uniquely identifies this storage bucket. E.g. we need to be able
        to disambiguate between an AWS S3 bucket called "foo" and a Google Cloud Storage
        bucket called "foo"
        """
        ...

    @abc.abstractmethod
    async def get_bytes(self, key: str) -> bytes:
        ...

    @abc.abstractmethod
    async def try_get_bytes(self, key: str) -> Optional[bytes]:
        ...

    async def get_bytes_and_key(self, key: str) -> Tuple[bytes, str]:
        """
        A thin wrapper around get_bytes that also returns the key. Wouldn't be necessary
        if asyncio was easier to work with
        """
        return await self.get_bytes(key), key

    @abc.abstractmethod
    async def get_byte_range(self, key: str, byte_range: Tuple[int, int]) -> bytes:
        ...

    @abc.abstractmethod
    async def write_bytes(self, data: bytes, key: str) -> None:
        ...

    @abc.abstractmethod
    async def exists(self, key: str) -> bool:
        ...

    async def write_bytes_if_not_exists(self, data: bytes, key: str) -> None:
        if not await self.exists(key):
            await self.write_bytes(data, key)

    @abc.abstractmethod
    async def get_file(self, key: str, local_filename: str) -> None:
        ...

    @abc.abstractmethod
    async def try_get_file(self, key: str, local_filename: str) -> bool:
        ...

    @abc.abstractmethod
    async def write_file(self, local_filename: str, key: str) -> None:
        ...

    @abc.abstractmethod
    async def list_objects(self, key_prefix: str) -> List[str]:
        ...

    @abc.abstractmethod
    async def delete_object(self, key: str) -> None:
        ...



from __future__ import annotations

import argparse
from typing import Any, Optional, Dict, List
import boto3

import meadowrun.aws_integration.management_lambdas.adjust_ec2_instances
import meadowrun.aws_integration.management_lambdas.clean_up
from meadowrun.aws_integration.aws_core import (
    _MEADOWRUN_USER_GROUP_NAME,
    _get_current_ip_for_ssh,
    get_bucket_name,
)
from meadowrun.aws_integration.aws_mgmt_lambda_install import (
    _CLEAN_UP_LAMBDA_NAME,
    _CLEAN_UP_LAMBDA_SCHEDULE_RULE,
    _EC2_ALLOC_LAMBDA_NAME,
    _EC2_ALLOC_LAMBDA_SCHEDULE_RULE,
    ensure_clean_up_lambda,
    ensure_ec2_alloc_lambda,
)
from meadowrun.aws_integration.aws_permissions_install import (
    _EC2_ROLE_INSTANCE_PROFILE,
    _EC2_ROLE_NAME,
    _MANAGEMENT_LAMBDA_POLICY_NAME,
    _MANAGEMENT_LAMBDA_ROLE,
    _EC2_ROLE_POLICY_NAME,
    _MEADOWRUN_USER_POLICY_NAME,
    _policy_arn_from_name,
    ensure_management_lambda_role,
    ensure_meadowrun_ec2_role,
    ensure_meadowrun_user_group,
)
from meadowrun.aws_integration.ec2 import (
    _MEADOWRUN_SSH_SECURITY_GROUP,
    ensure_security_group,
)
from meadowrun.aws_integration.ec2_instance_allocation import ensure_ec2_alloc_table
from meadowrun.aws_integration.ec2_pricing import clear_prices_cache
from meadowrun.aws_integration.ec2_ssh_keys import (
    MEADOWRUN_KEY_PAIR_NAME,
    _MEADOWRUN_KEY_PAIR_SECRET_NAME,
    ensure_meadowrun_key_pair,
)
from meadowrun.aws_integration.ecr import _ensure_repository
from meadowrun.aws_integration.management_lambdas.adjust_ec2_instances import (
    _get_non_terminated_instances,
)
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    _EC2_ALLOC_TABLE_NAME,
    _MEADOWRUN_GENERATED_DOCKER_REPO,
    ignore_boto3_error_code,
)


def add_management_lambda_override_arguments(parser: argparse.ArgumentParser) -> None:
    """Modifies parser in place!"""
    parser.add_argument("--terminate-instances-if-idle-for-secs")


def get_management_lambda_overrides_from_args(args: Any) -> Dict[str, str]:
    overrides = {}
    if args.terminate_instances_if_idle_for_secs:
        overrides[
            "TERMINATE_INSTANCES_IF_IDLE_FOR_SECS"
        ] = args.terminate_instances_if_idle_for_secs
    return overrides


async def install(
    region_name: str,
    allow_authorize_ips: bool,
    vpc_id: Optional[str],
    management_lambda_overrides: Dict[str, str],
) -> None:
    """Installs resources needed to run Meadowrun jobs"""

    ensure_security_group(_MEADOWRUN_SSH_SECURITY_GROUP, region_name, vpc_id)

    if not allow_authorize_ips:
        print(
            "--allow-authorize-ips was set to false. This means you must manually edit "
            f"the {_MEADOWRUN_SSH_SECURITY_GROUP} security group to allow users to SSH "
            "over port 22 to that security group. For example, you can run `aws ec2 "
            "authorize-security-group-ingress --group-name "
            f"{_MEADOWRUN_SSH_SECURITY_GROUP} --protocol tcp --port 22 --cidr "
            f"{await _get_current_ip_for_ssh()}/32`"
        )

    iam_client = boto3.client("iam", region_name=region_name)
    ensure_meadowrun_user_group(iam_client, allow_authorize_ips)
    ensure_meadowrun_ec2_role(iam_client)
    ensure_management_lambda_role(iam_client)

    ensure_ec2_alloc_table(region_name)

    await ensure_ec2_alloc_lambda(True, management_lambda_overrides)
    await ensure_clean_up_lambda(True, management_lambda_overrides)

    ensure_meadowrun_key_pair(region_name)

    ensure_bucket(region_name)

    _ensure_repository(_MEADOWRUN_GENERATED_DOCKER_REPO, region_name)


async def edit_management_lambda_config(overrides: Dict[str, str]) -> None:
    await ensure_ec2_alloc_lambda(True, overrides)
    await ensure_clean_up_lambda(True, overrides)


def _delete_user_group(iam: Any, group_name: str) -> None:
    """
    Deletes an IAM user group, requires removing all users and detaching all policies
    """
    success, group = ignore_boto3_error_code(
        lambda: iam.get_group(GroupName=group_name), "NoSuchEntity"
    )
    if success:
        assert group is not None  # just for mypy
        for user in group["Users"]:
            iam.remove_user_from_group(GroupName=group_name, UserName=user["UserName"])

    success, attached_policies = ignore_boto3_error_code(
        lambda: iam.list_attached_group_policies(GroupName=group_name), "NoSuchEntity"
    )
    if success:
        assert attached_policies is not None  # just for mypy
        for attached_policy in attached_policies["AttachedPolicies"]:
            iam.detach_group_policy(
                GroupName=group_name, PolicyArn=attached_policy["PolicyArn"]
            )

    success, inline_policies = ignore_boto3_error_code(
        lambda: iam.list_group_policies(GroupName=group_name), "NoSuchEntity"
    )
    if success:
        assert inline_policies is not None  # just for mypy
        for inline_policy in inline_policies["PolicyNames"]:
            iam.delete_group_policy(GroupName=group_name, PolicyName=inline_policy)

    ignore_boto3_error_code(
        lambda: iam.delete_group(GroupName=group_name), "NoSuchEntity"
    )


def _delete_iam_role(iam: Any, role_name: str) -> None:
    """Deletes an IAM role, requires detaching all policies first"""
    success, attached_policies = ignore_boto3_error_code(
        lambda: iam.list_attached_role_policies(RoleName=role_name), "NoSuchEntity"
    )
    if success:
        assert attached_policies is not None  # just for mypy
        for attached_policy in attached_policies["AttachedPolicies"]:
            iam.detach_role_policy(
                RoleName=role_name, PolicyArn=attached_policy["PolicyArn"]
            )

    success, inline_policies = ignore_boto3_error_code(
        lambda: iam.list_role_policies(RoleName=role_name), "NoSuchEntity"
    )
    if success:
        assert inline_policies is not None  # just for mypy
        for inline_policy in inline_policies["PolicyNames"]:
            iam.delete_role_policy(RoleName=role_name, PolicyName=inline_policy)

    ignore_boto3_error_code(lambda: iam.delete_role(RoleName=role_name), "NoSuchEntity")


def _delete_policy(iam: Any, policy_name: str) -> None:
    policy_arn = _policy_arn_from_name(policy_name)

    # non-default versions need to be deleted explicitly
    success, versions = ignore_boto3_error_code(
        lambda: iam.list_policy_versions(PolicyArn=policy_arn), "NoSuchEntity"
    )
    if success:
        assert versions is not None  # just for mypy
        for version in versions["Versions"]:
            if not version["IsDefaultVersion"]:
                iam.delete_policy_version(
                    PolicyArn=policy_arn, VersionId=version["VersionId"]
                )
    ignore_boto3_error_code(
        lambda: iam.delete_policy(PolicyArn=_policy_arn_from_name(policy_name)),
        "NoSuchEntity",
    )


def _delete_event_rule(events_client: Any, rule_name: str) -> None:
    """Deletes an EventBridge rule, requires deleting targets first"""
    success, target_ids = ignore_boto3_error_code(
        lambda: [
            target["Id"]
            for target in events_client.list_targets_by_rule(Rule=rule_name)["Targets"]
        ],
        "ResourceNotFoundException",
    )
    if success and target_ids:
        events_client.remove_targets(Rule=rule_name, Ids=target_ids)
    events_client.delete_rule(Name=rule_name)


def terminate_all_instances(region_name: str, wait: bool) -> None:
    """
    Terminates all Meadowrun-tagged instances, regardless of whether they are registered
    or not. If wait is true, we wait for the instances to terminate, otherwise we return
    "immediately". WARNING this will kill running jobs.
    """
    instances = _get_non_terminated_instances(
        boto3.resource("ec2", region_name=region_name)
    )
    for instance in instances:
        print(f"Terminating {instance.id}")
        instance.terminate()
    if wait:
        print("Waiting for instances to terminate cleanly")
        for instance in instances:
            instance.wait_until_terminated()


def delete_meadowrun_resources(region_name: str) -> None:
    """
    Delete all AWS resources that meadowrun creates

    This needs to contain all old names/types of resources in every published version of
    this library.
    """

    terminate_all_instances(region_name, True)

    meadowrun.aws_integration.management_lambdas.clean_up.delete_all_task_queues(
        region_name
    )

    delete_bucket(region_name)

    iam = boto3.client("iam", region_name=region_name)

    ignore_boto3_error_code(
        lambda: iam.remove_role_from_instance_profile(
            RoleName=_EC2_ROLE_NAME,
            InstanceProfileName=_EC2_ROLE_INSTANCE_PROFILE,
        ),
        "NoSuchEntity",
    )
    ignore_boto3_error_code(
        lambda: iam.delete_instance_profile(
            InstanceProfileName=_EC2_ROLE_INSTANCE_PROFILE
        ),
        "NoSuchEntity",
    )

    _delete_user_group(iam, _MEADOWRUN_USER_GROUP_NAME)
    _delete_iam_role(iam, _EC2_ROLE_NAME)
    _delete_iam_role(iam, _MANAGEMENT_LAMBDA_ROLE)

    _delete_policy(iam, _MEADOWRUN_USER_POLICY_NAME)
    _delete_policy(iam, _EC2_ROLE_POLICY_NAME)
    _delete_policy(iam, _MANAGEMENT_LAMBDA_POLICY_NAME)

    lambda_client = boto3.client("lambda", region_name=region_name)
    ignore_boto3_error_code(
        lambda: lambda_client.delete_function(FunctionName=_EC2_ALLOC_LAMBDA_NAME),
        "ResourceNotFoundException",
    )
    ignore_boto3_error_code(
        lambda: lambda_client.delete_function(FunctionName=_CLEAN_UP_LAMBDA_NAME),
        "ResourceNotFoundException",
    )

    logs_client = boto3.client("logs", region_name=region_name)
    ignore_boto3_error_code(
        lambda: logs_client.delete_log_group(
            logGroupName=f"/aws/lambda/{_EC2_ALLOC_LAMBDA_NAME}"
        ),
        "ResourceNotFoundException",
    )
    ignore_boto3_error_code(
        lambda: logs_client.delete_log_group(
            logGroupName=f"/aws/lambda/{_CLEAN_UP_LAMBDA_NAME}"
        ),
        "ResourceNotFoundException",
    )

    events_client = boto3.client("events", region_name=region_name)
    _delete_event_rule(events_client, _EC2_ALLOC_LAMBDA_SCHEDULE_RULE)
    _delete_event_rule(events_client, _CLEAN_UP_LAMBDA_SCHEDULE_RULE)

    ec2_client = boto3.client("ec2", region_name=region_name)
    ec2_client.delete_key_pair(KeyName=MEADOWRUN_KEY_PAIR_NAME)
    # TODO this will fail if there are unterminated instances using this security group.
    # We terminate all instances at the beginning of this function, but it takes time to
    # terminate them.
    ignore_boto3_error_code(
        lambda: ec2_client.delete_security_group(
            GroupName=_MEADOWRUN_SSH_SECURITY_GROUP
        ),
        "InvalidGroup.NotFound",
    )

    secrets_client = boto3.client("secretsmanager", region_name=region_name)
    ignore_boto3_error_code(
        lambda: secrets_client.delete_secret(SecretId=_MEADOWRUN_KEY_PAIR_SECRET_NAME),
        "ResourceNotFoundException",
    )

    dynamodb_client = boto3.client("dynamodb", region_name=region_name)
    ignore_boto3_error_code(
        lambda: dynamodb_client.delete_table(TableName=_EC2_ALLOC_TABLE_NAME),
        "ResourceNotFoundException",
    )

    ecr_client = boto3.client("ecr", region_name=region_name)
    ignore_boto3_error_code(
        lambda: ecr_client.delete_repository(
            repositoryName=_MEADOWRUN_GENERATED_DOCKER_REPO, force=True
        ),
        "RepositoryNotFoundException",
    )

    clear_prices_cache()


def delete_bucket(region_name: str) -> None:
    """Deletes the meadowrun bucket"""
    s3 = boto3.resource("s3", region_name=region_name)

    bucket = s3.Bucket(get_bucket_name(region_name))
    success, _ = ignore_boto3_error_code(
        lambda: list(bucket.objects.limit(1)),
        "NoSuchBucket",
    )
    if success:
        # S3 doesn't allow deleting a bucket with anything in it, so delete all objects
        # in chunks of up to 1000, which is the maximum allowed.
        key_chunk: List[Dict[str, str]] = []
        for s3object in bucket.objects.all():
            if len(key_chunk) == 1000:
                bucket.delete_objects(Delete=dict(Objects=key_chunk))
                key_chunk.clear()
            key_chunk.append(dict(Key=s3object.key))
        if key_chunk:
            bucket.delete_objects(Delete=dict(Objects=key_chunk))

        bucket.delete()


def ensure_bucket(
    region_name: str,
    expire_days: int = 14,
) -> None:
    """
    Create an S3 bucket in a specified region if it does not exist yet.

    If a region is not specified, the bucket is created in the configured default
    region.

    The bucket is created with a default lifecycle policy of 14 days.

    Since bucket names must be globally unique, the name is bucket_prefix + region +
    account number

    :param bucket_name: Bucket to create
    :param region_name: String region to create bucket in, e.g., 'us-west-2'
    :param expire_days: int number of days after which keys are deleted by lifecycle
        policy.
    :return: the full bucket name
    """

    s3 = boto3.client("s3", region_name=region_name)

    bucket_name = get_bucket_name(region_name)

    # us-east-1 cannot be specified as a LocationConstraint because it is the default
    # region
    # https://stackoverflow.com/questions/51912072/invalidlocationconstraint-error-while-creating-s3-bucket-when-the-used-command-i
    if region_name == "us-east-1":
        additional_parameters = {}
    else:
        additional_parameters = {
            "CreateBucketConfiguration": {"LocationConstraint": region_name}
        }

    success, _ = ignore_boto3_error_code(
        lambda: s3.create_bucket(Bucket=bucket_name, **additional_parameters),
        "BucketAlreadyOwnedByYou",
    )
    s3.put_bucket_lifecycle_configuration(
        Bucket=bucket_name,
        LifecycleConfiguration=dict(
            Rules=[
                dict(
                    Expiration=dict(
                        Days=expire_days,
                    ),
                    ID="meadowrun-lifecycle-policy",
                    # Filter is mandatory, but we don't want one:
                    Filter=dict(Prefix=""),
                    Status="Enabled",
                )
            ]
        ),
    )



"""
This module contains functionality for interacting with Kubernetes that theoretically
has nothing to do with Meadowrun. I.e. you can think of this as a slightly higher-level
version of the kubernetes_asyncio module.
"""

from __future__ import annotations

import asyncio
import base64
import enum
import time
from typing import Callable, List, Dict, Tuple, Optional

from kubernetes_asyncio import client as kubernetes_client, watch as kubernetes_watch

from meadowrun.run_job_core import WaitOption


# kubernetes_asyncio.watch.Watch looks at the docstring to figure out if the function
# passed to Watch takes a follow or watch argument. This results in the wrong value for
# read_namespaced_pod_log, so we just monkey-patch the function here
_orig_get_watch_argument_name = kubernetes_watch.Watch.get_watch_argument_name


def _new_get_watch_argument_name(watch: kubernetes_watch.Watch, func: Callable) -> str:
    if getattr(func, "__name__") == "read_namespaced_pod_log":
        return "follow"
    return _orig_get_watch_argument_name(watch, func)


kubernetes_watch.watch.Watch.get_watch_argument_name = _new_get_watch_argument_name


async def get_pods_for_job(
    core_api: kubernetes_client.CoreV1Api,
    kubernetes_namespace: str,
    job_id: str,
    pod_generate_names: List[str],
) -> List[kubernetes_client.V1Pod]:
    """
    When you launch a Kubernetes job, one or more pods get created. In our case, we
    should only ever get one pod for regular jobs, and one pod for each index in indexed
    completion jobs, because we always set parallelism = completions, and we've
    configured 0 retries.

    For regular jobs, the pod created will be named <job_id>-<random string>. The pod's
    metadata has a generate_name, which will be equal to <job_id>-.

    For indexed completion jobs, there will be a pod for each index with a generate_name
    of <job_id>-<index>-.

    pod_generate_names should have the list of generate_names that we expect to see for
    the specified job.

    Returns a pod object corresponding to each pod_generate_names, in the same order as
    pod_generate_names.
    """

    pod_generate_names_set = set(pod_generate_names)
    results: Dict[str, kubernetes_client.V1Pod] = {}
    i = 0

    while True:
        pods = await core_api.list_namespaced_pod(
            kubernetes_namespace, label_selector=f"job-name={job_id}"
        )
        for pod in pods.items:
            generate_name = pod.metadata.generate_name

            if generate_name not in pod_generate_names_set:
                raise ValueError(
                    f"Unexpected pod {pod.metadata.name} with generate name "
                    f"{generate_name} found"
                )

            if (
                generate_name in results
                and results[generate_name].metadata.name != pod.metadata.name
            ):
                # TODO we may need to change this if we add e.g. retries
                raise ValueError(
                    "Unexpected multiple pods with the same generate name "
                    f"{generate_name} found: {results[generate_name].metadata.name}, "
                    f"{pod.metadata.name}"
                )

            results[generate_name] = pod

        if len(results) >= len(pod_generate_names):
            break

        if i > 15:
            raise TimeoutError(
                "Waited >15s, but pods with the following generate names were not "
                "created: "
                + ", ".join(p for p in pod_generate_names if p not in results)
            )

        if i == 0:
            print(f"Waiting for pods to be created for the job {job_id}")

        await asyncio.sleep(1.0)

        i += 1

    return [results[generate_name] for generate_name in pod_generate_names]


def get_main_container_state(
    pod: kubernetes_client.V1Pod, job_id: str
) -> Tuple[Optional[kubernetes_client.V1ContainerState], Optional[str]]:
    # first get main container state
    container_statuses = pod.status.container_statuses

    if container_statuses is None or len(container_statuses) == 0:
        main_container_state = None
    else:
        main_container_statuses = [s for s in container_statuses if s.name == "main"]
        if len(main_container_statuses) == 0:
            raise ValueError(
                f"The job {job_id} has a pod {pod.metadata.name} but there is no `main`"
                " container"
            )
        if len(main_container_statuses) > 1:
            raise ValueError(
                f"The job {job_id} has a pod {pod.metadata.name} but there is more than"
                " one `main` container"
            )
        main_container_state = main_container_statuses[0].state

    # then get the latest condition's reason, this is where Kubernetes will tell us that
    # e.g. the pod is unschedulable
    if pod.status.conditions:
        latest_condition = pod.status.conditions[-1]
        result_builder = []
        if latest_condition.reason:
            result_builder.append(latest_condition.reason)
        if latest_condition.message:
            result_builder.append(latest_condition.message)
        latest_condition_reason = ", ".join(result_builder)
    else:
        latest_condition_reason = None

    return main_container_state, latest_condition_reason


# This is how long we'll wait for a pod that is pulling an image
_WAIT_FOR_POD_PULLING_IMAGE_SECS = 60 * 7
# This is how long we'll wait for a pod to that isn't running yet and is in the
# "unschedulable" state. We want to allow some time for a pod autoscaler to kick in
_WAIT_FOR_POD_UNSCHEDULABLE_SECS = 60 * 7
# This is how long we'll wait for a pod to start running for any state other than the
# ones described above
_WAIT_FOR_POD_RUNNING_OTHER_SECS = 15


class PodState(enum.Enum):
    PULLING_IMAGE = 1
    UNSCHEDULABLE = 2
    OTHER_HAS_NOT_RUN = 3
    STARTED_RUNNING = 4


def _get_pod_state(pod: kubernetes_client.V1Pod, job_id: str) -> Tuple[PodState, str]:
    """
    Returns the internal representation of the pod state (PodState) and a string to
    present to the user about the state of the pod
    """
    main_container_state, latest_condition_reason = get_main_container_state(
        pod, job_id
    )
    if main_container_state is not None and (
        main_container_state.running is not None
        or main_container_state.terminated is not None
    ):
        return PodState.STARTED_RUNNING, ""

    pod_state = PodState.OTHER_HAS_NOT_RUN

    additional_info_builder = [":"]
    if main_container_state is not None and main_container_state.waiting is not None:
        waiting_reason = str(main_container_state.waiting.reason)
        additional_info_builder.append(waiting_reason)

        if main_container_state.waiting.message is not None:
            additional_info_builder.append(str(main_container_state.waiting.message))
        elif (
            main_container_state.waiting.reason == "ContainerCreating"
            and pod_state == PodState.OTHER_HAS_NOT_RUN
        ):
            # TODO this isn't exactly right, there's a separate Event (different API)
            # that would tell us conclusively that we're waiting for the image pull, but
            # this logic seems to be correct most of the time. We should maybe read
            # those events, but at some point we may get what we need in this field
            # waiting.reason field:
            # https://github.com/kubernetes/kubernetes/issues/19077
            pod_state = PodState.PULLING_IMAGE
            additional_info_builder.append("(pulling image)")
    if latest_condition_reason:
        additional_info_builder.append(latest_condition_reason)
        if latest_condition_reason.startswith("Unschedulable"):
            pod_state = PodState.UNSCHEDULABLE

    if len(additional_info_builder) == 1:
        additional_info = ""
    else:
        additional_info = " ".join(additional_info_builder)

    return pod_state, additional_info


async def wait_for_pod_running(
    core_api: kubernetes_client.CoreV1Api,
    job_id: str,
    kubernetes_namespace: str,
    pod: kubernetes_client.V1Pod,
) -> kubernetes_client.V1Pod:
    pod_name = pod.metadata.name

    # The first step is to wait for the pod to start running, because we can't stream
    # logs until the pod is in a running state. The happy path is that our pod is in the
    # "waiting" state because we're either waiting for the image to get pulled or we're
    # waiting for nodes to become available to run our job. In that case, we'll wait up
    # to 7 minutes.
    #
    # The unhappy path is that something has gone wrong which Kubernetes expresses as
    # waiting infinitely, rather than a failure. E.g. if our image spec is invalid. In
    # that case we'll only wait 15 seconds, as it doesn't make sense to expect that that
    # would change.

    seconds_waited = 0
    prev_additional_info = None
    prev_pod_state = PodState.OTHER_HAS_NOT_RUN
    wait_until = _WAIT_FOR_POD_RUNNING_OTHER_SECS
    pod_state, additional_info = _get_pod_state(pod, job_id)
    while pod_state != PodState.STARTED_RUNNING:
        if additional_info != prev_additional_info:
            print(f"Waiting for pod {pod_name} to start running{additional_info}")
            prev_additional_info = additional_info

        await asyncio.sleep(1.0)  # this is the polling interval
        seconds_waited += 1
        if pod_state != prev_pod_state:
            # every time we "change state", extend/shorten how much time we will wait
            # based on the new state
            prev_pod_state = prev_pod_state
            if pod_state == PodState.PULLING_IMAGE:
                wait_until = seconds_waited + _WAIT_FOR_POD_PULLING_IMAGE_SECS
            elif pod_state == PodState.UNSCHEDULABLE:
                wait_until = seconds_waited + _WAIT_FOR_POD_UNSCHEDULABLE_SECS
            else:
                wait_until = seconds_waited + _WAIT_FOR_POD_RUNNING_OTHER_SECS

        if seconds_waited > wait_until:
            raise TimeoutError(
                f"Waited >{seconds_waited} seconds for the container of job {job_id} in"
                " pod {pod_name} to start running"
            )

        pod = await core_api.read_namespaced_pod_status(pod_name, kubernetes_namespace)
        pod_state, additional_info = _get_pod_state(pod, job_id)

    return pod


async def stream_pod_logs(
    core_api: kubernetes_client.CoreV1Api, kubernetes_namespace: str, pod_name: str
) -> None:
    # Now our pod is running, so we can stream the logs

    async with kubernetes_watch.Watch() as w:
        async for line in w.stream(
            core_api.read_namespaced_pod_log,
            name=pod_name,
            namespace=kubernetes_namespace,
        ):
            print(line, end="")


async def wait_for_pod_exit(
    core_api: kubernetes_client.CoreV1Api,
    job_id: str,
    kubernetes_namespace: str,
    pod_name: str,
    timeout_seconds: int,
    streamed_logs: bool,
) -> int:
    # Once this stream ends, we know the pod is completed, but sometimes it takes some
    # time for Kubernetes to report that the pod has completed. So we poll until the pod
    # is reported as terminated.

    pod = await core_api.read_namespaced_pod_status(pod_name, kubernetes_namespace)
    main_container_state, _ = get_main_container_state(pod, job_id)
    t0 = time.time()
    while main_container_state is None or main_container_state.running is not None:
        await asyncio.sleep(1.0)
        if time.time() > t0 + timeout_seconds:
            if streamed_logs:
                raise TimeoutError(
                    f"Unexpected. The job {job_id} has a pod {pod_name}, and the pod "
                    f"still seems to be running {timeout_seconds} seconds after the log"
                    " stream ended"
                )
            else:
                raise TimeoutError(
                    f"The job {job_id} timed out, the pod {pod_name} as been running "
                    f"for {timeout_seconds} seconds"
                )
        pod = await core_api.read_namespaced_pod_status(pod_name, kubernetes_namespace)
        main_container_state, _ = get_main_container_state(pod, job_id)

    return main_container_state.terminated.exit_code


async def wait_for_pod(
    core_api: kubernetes_client.CoreV1Api,
    job_id: str,
    kubernetes_namespace: str,
    pod: kubernetes_client.V1Pod,
    wait_for_result: WaitOption,
) -> int:
    """
    This function waits for the specified pod to start running, streams the logs from
    that pod into our local stdout, and then waits for the pod to terminate. Then we
    return the exit code of the pod.
    """
    await wait_for_pod_running(core_api, job_id, kubernetes_namespace, pod)
    if wait_for_result == WaitOption.DO_NOT_WAIT:
        # TODO maybe return None instead? Currently this code path is not used, requires
        # support in the caller
        return 0

    if wait_for_result == WaitOption.WAIT_AND_TAIL_STDOUT:
        await stream_pod_logs(core_api, kubernetes_namespace, pod.metadata.name)

        return await wait_for_pod_exit(
            core_api, job_id, kubernetes_namespace, pod.metadata.name, 15, True
        )
    else:
        # TODO this timeout should be configurable and the default should be smaller
        # than 2 days
        wait_for_pod_exit_timeout_seconds = 60 * 60 * 24 * 2
        return await wait_for_pod_exit(
            core_api,
            job_id,
            kubernetes_namespace,
            pod.metadata.name,
            wait_for_pod_exit_timeout_seconds,
            False,
        )


async def get_kubernetes_secret(
    kubernetes_namespace: str, secret_name: str
) -> Dict[str, str]:
    async with kubernetes_client.ApiClient() as api_client:
        core_api = kubernetes_client.CoreV1Api(api_client)
        result = await core_api.read_namespaced_secret(
            secret_name, kubernetes_namespace
        )

    return {
        key: base64.b64decode(value).decode("utf-8")
        for key, value in result.data.items()
    }


async def run_command_on_pod(
    pod_name: str,
    kubernetes_namespace: str,
    command: List[str],
    ws_core_api: kubernetes_client.CoreV1Api,
) -> str:
    """Returns the output of running command"""
    return await ws_core_api.connect_post_namespaced_pod_exec(
        name=pod_name,
        namespace=kubernetes_namespace,
        command=command,
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
    )


def get_main_container_is_ready(pod: kubernetes_client.V1Pod) -> bool:
    main_container_statuses = [
        container
        for container in pod.status.container_statuses
        if container.name == "main"
    ]
    if len(main_container_statuses) == 0:
        print(f"Unexpected: pod {pod.metadata.name} has no main container, skipping")
        return False
    elif len(main_container_statuses) > 1:
        print(
            f"Unexpected: pod {pod.metadata.name} has more than one main container, "
            "skipping"
        )
        return False
    else:
        return main_container_statuses[0].ready


async def set_main_container_ready(
    core_api: kubernetes_client.CoreV1Api, pod: kubernetes_client.V1Pod, is_ready: bool
) -> None:
    main_containers = [
        container
        for container in pod.status.container_statuses
        if container.name == "main"
    ]
    if len(main_containers) != 1:
        raise ValueError(
            "Unexpected number of container with the name 'main': "
            f"{len(main_containers)}"
        )

    main_container_state = main_containers[0]
    main_container_state.ready = is_ready

    await core_api.patch_namespaced_pod_status(
        pod.metadata.name,
        pod.metadata.namespace,
        kubernetes_client.V1Pod(
            status=kubernetes_client.V1PodStatus(
                container_statuses=[main_container_state]
            )
        ),
    )



import asyncssh

import boto3

from meadowrun.aws_integration.aws_core import wrap_access_or_install_errors
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    ignore_boto3_error_code,
)

MEADOWRUN_KEY_PAIR_NAME = "meadowrun_key_pair"
_MEADOWRUN_KEY_PAIR_SECRET_NAME = "meadowrun_private_key"


def ensure_meadowrun_key_pair(region_name: str) -> None:
    """
    Ensures that an EC2 key pair and corresponding private key in an AWS secret exist.
    If both exist, they will not be modified. If either do not exist, both will be
    recreated.
    """

    ec2_client = boto3.client("ec2", region_name=region_name)
    secrets_client = boto3.client("secretsmanager", region_name=region_name)

    # first figure out the state of the secret, and if it has been deleted, restore it

    success, secret_value_result, error_code = ignore_boto3_error_code(
        lambda: secrets_client.get_secret_value(
            SecretId=_MEADOWRUN_KEY_PAIR_SECRET_NAME
        ),
        {
            "ResourceNotFoundException",
            "InvalidRequestException",
        },
        True,
    )

    if success:
        secret_state = "exists"
    elif error_code == "InvalidRequestException":
        # this means the secret is marked for deletion, so we need to restore it then
        # overwrite it
        secrets_client.restore_secret(SecretId=_MEADOWRUN_KEY_PAIR_SECRET_NAME)
        secret_state = "restored"
    elif error_code == "ResourceNotFoundException":
        # this means the secret doesn't exist
        secret_state = "does not exist"
    else:
        raise ValueError(
            f"success was not True, but error_code was an unexpected value {error_code}"
        )

    # next figure out the state of the key pair

    key_pair_exists, _ = ignore_boto3_error_code(
        lambda: ec2_client.describe_key_pairs(KeyNames=[MEADOWRUN_KEY_PAIR_NAME]),
        "InvalidKeyPair.NotFound",
    )

    if secret_state in ("restored", "does not exist") or not key_pair_exists:
        ec2_client.delete_key_pair(KeyName=MEADOWRUN_KEY_PAIR_NAME)
        create_key_pair_result = ec2_client.create_key_pair(
            KeyName=MEADOWRUN_KEY_PAIR_NAME, KeyType="rsa"
        )
        private_key_text = create_key_pair_result["KeyMaterial"]
        if secret_state == "does not exist":
            secrets_client.create_secret(
                Name=_MEADOWRUN_KEY_PAIR_SECRET_NAME, SecretString=private_key_text
            )
        else:  # restored or exists
            secrets_client.put_secret_value(
                SecretId=_MEADOWRUN_KEY_PAIR_SECRET_NAME, SecretString=private_key_text
            )


def _get_meadowrun_ssh_key_text(region_name: str) -> str:
    # TODO should we cache the key locally so that we're not requesting the secret
    # constantly?
    secrets_client = boto3.client("secretsmanager", region_name=region_name)
    secret_result = wrap_access_or_install_errors(
        lambda: secrets_client.get_secret_value(
            SecretId=_MEADOWRUN_KEY_PAIR_SECRET_NAME
        ),
        f"secret {_MEADOWRUN_KEY_PAIR_SECRET_NAME}",
        "AccessDeniedException",
        # InvalidRequestException actually means this was deleted
        {"ResourceNotFoundException", "InvalidRequestException"},
    )
    return secret_result["SecretString"]


def get_meadowrun_ssh_key(region_name: str) -> asyncssh.SSHKey:
    """
    Returns an asyncssh.SSHKey that is compatible with that EC2 key pair and can be
    passed to ssh.connect(private_key=pkey).
    """
    private_key_text = _get_meadowrun_ssh_key_text(region_name)
    return asyncssh.import_private_key(private_key_text)


def download_ssh_key(output_path: str, region_name: str) -> None:
    private_key_text = _get_meadowrun_ssh_key_text(region_name)
    with open(output_path, "w", encoding="utf-8") as output_file:
        output_file.write(private_key_text)
